/* 
 * x-tlk Toolkit - Copyright (c) Holger Pfaff
 * http://pfaff.dnsalias.net
 *
 * This software maybe used for any purpose provided the
 * above copyright notice is retained. It is supplied as is.
 * No warranty expressed or implied - Use at your own risk.
 */

#ifndef _But_c
#define _But_c
#define _Tlk_

#include <tlk.h>

void But_init() {
   But_arrow = XCreatePixmapFromBitmapData(Dsp, Dsp_root,
                  But_arrow_bits, BUT_ARROW_WIDTH, BUT_ARROW_HEIGHT,
                  Dsp_black, Dsp_color ? Dsp_grey50 : Dsp_white, Dsp_depth);
}


But But_create(Win win, int row, int col, char *label, Mnu mnu, void (*proc)()) {
   But but = Win_create(win, WIN_BUT_WIN, row, col, 1, strlen(label) + (mnu ? 3 : 2), win->scale, NULL, label, proc);
   if(mnu) {
      Win_setflag(but, WIN_F_HAS_SUBMENU);
      but->data = (char *) mnu;
   }
   return but;
}

void But_3d(But but, GC l, GC r, GC m) {
   Win_3d(but->window, l, r, m, 2, 2, but->cols * Dsp_rw[but->scale] - 5, but->rows * Dsp_rh[but->scale] + 1);
}

void But_label(But but) {
   char *ptr = but->label;
   Mni mni;
   
   if(but->type == WIN_BUT_WIN && Win_getflag(but, WIN_F_HAS_SUBMENU) && Mnu_act == -1) {
      mni = (Mni) Tlk_get((Win)(but->data), TLK_GET_MNI, (caddr_t) 0);
      ptr = mni->label;
   }
   
   Win_uml[0] = '\0';
   if(ptr) strcpy(Win_uml, ptr);
   Win_uml[but->cols - (but->type==WIN_MNI_WIN ? 1 : 2)] = '\0';
}

void But_draw(But but) {
   if(Dsp_color && but->type != WIN_MNI_WIN) {
      But_3d(but, Dspgc_25, Dspgc_75, Dspgc_50); 
   } else {
      XClearWindow(Dsp, but->window);
   }
         
   if(but->label == 0) return;

   Win_text(but, Win_getflag(but, WIN_F_DEACTIVATED) ? Dspgc_75 : Dspgc, 
            0, 1, -1, Dsp_color && but->type != WIN_MNI_WIN ? 3 : 0, 
            but->label, strlen(but->label), but->type == WIN_MNI_WIN ? 1 : 2);

   if(Win_getflag(but, WIN_F_HAS_SUBMENU) &&
     !Win_getflag(but, WIN_F_DEACTIVATED)   ) {
      XCopyArea(Dsp, But_arrow, but->window, Dspgc, 0, 0,
                BUT_ARROW_WIDTH, BUT_ARROW_HEIGHT,
                but->cols * Dsp_rw[but->scale] - BUT_ARROW_WIDTH - 4,
               (Dsp_rh[but->scale] / 2 - BUT_ARROW_HEIGHT / 2) + 
               (Dsp_color && but->type != WIN_MNI_WIN ? 3 : 0));
   }
}

void But_select(But but) {
   if(Dsp_color && but->type != WIN_MNI_WIN) {
      But_3d(but, Dspgc, Dspgc_25, Dspgc_75); 
   } else { 
      Win_fb(but, Dsp_color ? Dspgc_75 : Dspgc, 0, 0, 1, but->cols, 0, 0);
   }
   
   But_label(but);

   if(but->type == WIN_MNI_WIN)
      Win_ds(but, Dsp_color ? Dspgc : Dspgc_clr, 0, 1, -1, 0, Win_uml);
   else
      Win_bs(but, Dsp_color ? Dspgc : Dspgc_clr, 0, 1, Dsp_color ? 0 : -1, Dsp_color ? 4 : 0, Win_uml);

             
   if(Win_getflag(but, WIN_F_HAS_SUBMENU) && Mnu_act >= 0) {
      Win_dt(but, Dsp_color ? Dspgc : Dspgc_clr, 0, but->cols - 1, -1, 
             Dsp_color && but->type != WIN_MNI_WIN ? 3 : 0, ">", 1);
   }
}

void But_busy(But but) {
   if(Dsp_slow || but->type == WIN_MNI_WIN) return;
      
   if(Dsp_color) {
      But_3d(but, Dspgc_clr, Dspgc_75, Dspgc_25); 
   } else {
      Win_fb(but, Dspgc_25, 0, 0, 1, but->cols, 0, 0);
   }
   
   But_label(but);
   Win_bs(but, Dspgc_50, 0, 1, -1, Dsp_color ? 3 : 0, Win_uml);
   XFlush(Dsp);               
}

#endif /* _But_c */
