/* 
 * x-tlk Toolkit - Copyright (c) Holger Pfaff
 * http://pfaff.dnsalias.net
 *
 * This software maybe used for any purpose provided the
 * above copyright notice is retained. It is supplied as is.
 * No warranty expressed or implied - Use at your own risk.
 */

#ifndef _Dbk_c
#define _Dbk_c
#define _Tlk_

#include <tlk.h>
#include <sys/wait.h>
#include <unistd.h>

Dbk Dbk_create(int recl, int keyo, int keyl, void (*proc)()) {
   Dbk dbk = (Dbk) Win_malloc(sizeof(Dbk_s));
   dbk->recl  = recl;
   dbk->keyo  = keyo;
   dbk->keyl  = keyl;   
   dbk->proc  = proc;   
   dbk->recc  = 0;   
   dbk->flags = 0;   
   dbk->datap = 0;
   dbk->datas = 0;
   dbk->chpid = 0;
   dbk->filen = 0;
            
   return dbk;
}

void Dbk_destroy(Dbk dbk) {  
   if(dbk->datap) free(dbk->datap);
   if(dbk->filen) free(dbk->filen);
   if(dbk->chpid) waitpid(dbk->chpid, 0, 0);
   free(dbk);
}

int Dbk_load(Dbk dbk, char *filen) {
   int   size, fd;
   char *mod = "Dbk_load";
   
   Dbk_chkch(dbk);

   if(filen) {
      dbk->filen = (char *) Win_realloc(dbk->filen, strlen(filen));
      strcpy(dbk->filen, filen);
   }
   
   if(dbk->filen == 0) {
      Dbk_errproc(dbk, mod, Dbk_err_nofile);
      return -1;   
   }
      
   fd = open(dbk->filen, O_RDONLY);
   
   if(fd == -1) {
      Dbk_errproc(dbk, mod, "[1] open(dbk->filen, O_RDONLY)");
      return -1;
   }
   
   size = (int) lseek(fd, 0, 2);
   
   if(size % dbk->recl) {
      Dbk_errproc(dbk, mod, "[0] Incorrect filesize.");
      close(fd);
      return -1;
   }
      
   if(size == -1) {
      size = 0;
   }  
   
   dbk->datap  = (char *) Win_realloc(dbk->datap, dbk->datas + size);
   
   lseek(fd, 0, 0);
   read(fd, dbk->datap + dbk->datas, size);
   close(fd);
      
   dbk->datas += size;
   dbk->recc   = dbk->datas / dbk->recl;
   dbk->mtime  = Dbk_mtime(dbk->filen);
   
   return 0;     
}

Dbk Dbk_update(Dbk dbk, char *filen) {
   Dbk   dbknew;
   long  mtime;
   char *mod = "Dbk_update";
   
   Dbk_chkch(dbk);

   if(filen) {
      dbk->filen = (char *) Win_realloc(dbk->filen, strlen(filen));
      strcpy(dbk->filen, filen);
   }
   
   if(dbk->filen == 0) {
      Dbk_errproc(dbk, mod, Dbk_err_nofile);
      return dbk;   
   }

   mtime = Dbk_mtime(dbk->filen);

   if(mtime > dbk->mtime) {
      dbknew = Dbk_create(dbk->recl, dbk->keyo, dbk->keyl, dbk->proc);
      Dbk_load(dbknew, dbk->filen);
      dbknew->flags = dbk->flags;
      Dbk_destroy(dbk);
      dbk = dbknew;
   }
   return dbk;
}

int Dbk_write(Dbk dbk) {      
   int fd;
   int retc = 0;
   char *mod = "Dbk_write";
   
   if(Dbk_getflag(dbk, DBK_F_NOFORK) == FALSE) dbk->proc = 0;
   
   sprintf(Win_io, "%s.%ld", dbk->filen, time(0));
   if(rename(dbk->filen, Win_io) == -1) {
      Dbk_errproc(dbk, mod, "[1] rename(dbk->filen)");
      retc = 1;
      goto Dbk_write_exit;
   }
   fd = open(dbk->filen, O_WRONLY|O_TRUNC|O_CREAT, 0000660);
   if(fd == -1) {
      Dbk_errproc(dbk, mod, "[1] open(dbk->filen, O_WRONLY|O_TRUNC|O_CREAT, 0000660)");
      retc = 1;
      goto Dbk_write_exit;
   }
   if(write(fd, dbk->datap, dbk->datas) == -1) {
      Dbk_errproc(dbk, mod, "[1] write(fd, dbk->datap, dbk->datas)");
      retc = 1;
   }
#ifndef i386
   fsync(fd);
#endif
   if(close(fd) == -1) {
      Dbk_errproc(dbk, mod, "[1] close(fd)");
      retc = 1;
   }
   Dbk_delobsolete(dbk);

Dbk_write_exit:   
   sprintf(Win_io, "%s.lck", dbk->filen);
   if(unlink(Win_io) == -1) {
      Dbk_errproc(dbk, Win_io, "[1] unlink(.lck)");
      retc = 1;
   }
   Dbk_unsetflag(dbk, DBK_F_RDONLY);
   return retc;
}

int Dbk_save(Dbk dbk, char *filen) {
   int fd;
   char *mod = "Dbk_save";
   
   Dbk_chkch(dbk);
   
   if(filen) {
      dbk->filen = (char *) Win_realloc(dbk->filen, strlen(filen));
      strcpy(dbk->filen, filen);
   }
   
   if(dbk->filen == 0) {
      Dbk_errproc(dbk, mod, Dbk_err_nofile);
      return -1;   
   }
      
   if(Dbk_getflag(dbk, DBK_F_EDITED)) 
      Dbk_sort(dbk);
      
   sprintf(Win_io, "%s.lck", dbk->filen);
   while(access(Win_io, R_OK) == 0) sleep(1);

   if(access(dbk->filen, R_OK) == -1) {
      Dbk_errproc(dbk, mod, "[1] access(dbk->filen, R_OK)");
      return 1;
   }
      
   fd = open(Win_io, O_WRONLY|O_TRUNC|O_CREAT, 0000666);
   if(fd == -1) {
      Dbk_errproc(dbk, mod, "[1] open(.lck, O_WRONLY|O_TRUNC|O_CREAT, 0000666)");
      return 1;
   }
   close(fd);

   Dbk_setflag(dbk, DBK_F_RDONLY);
   
   if(Dbk_getflag(dbk, DBK_F_NOFORK) == TRUE) {
      return Dbk_write(dbk);
   }
   
   if((dbk->chpid = fork()) == 0) {
      Dbk_write(dbk);
      exit(0);
   }
   if(dbk->chpid == -1) {
      Dbk_errproc(dbk, mod, "[1] dbk->chpid = fork()");
      return -1;
   }
   return 0;
}   

int Dbk_sortkey(Dbk dbk, int off, int len) {
   Dbk_off   =  off;
   Dbk_len   =  len;
   Dbk_rev   =  Dbk_getflag(dbk, DBK_F_SORTR) ? -1 : 1;
   
   Dbk_chkch(dbk);
   
   qsort(dbk->datap, dbk->recc, dbk->recl, Dbk_cmp);
   
   if(Dbk_off == dbk->keyo && 
      Dbk_len == dbk->keyl && 
      Dbk_getflag(dbk, DBK_F_SORTR) == 0) {
      Dbk_unsetflag(dbk, DBK_F_EDITED);
   } else {
      Dbk_setflag(dbk, DBK_F_EDITED);
   }
   return 0;
}

int Dbk_put(Dbk dbk, char *rec, Dbk_put_t flag) {
   char *orec, *mod = "Dbk_put";
   Dbk_chkch(dbk);

   if(dbk->chpid) waitpid(dbk->chpid, 0, 0);

   if(flag != DBK_MULTIPLE) 
      orec  = Dbk_get(dbk, rec + dbk->keyo);
   
   switch(flag) {
      case DBK_REPLACE:
            if(orec) {
               Dbk_set(dbk, orec, rec);
               break;
            }
      case DBK_INSERT:
            if(orec) {
               Dbk_errproc(dbk, mod, "[1] Insert: Key exists.");
               return -1;
            }
      case DBK_MULTIPLE:
            Dbk_new(dbk, rec);
            break;
   }
   return 0;
}

char *Dbk_getrec(Dbk dbk, int no) {
   return (no < dbk->recc) ? dbk->datap + (dbk->recl * no) : NULL;
}

char *Dbk_getkey(Dbk dbk, int off, int len, char *key) {      
   register int   recc = dbk->recc;
   register int   recl = dbk->recl, cmp;
   register char *base = dbk->datap + off, *nbase;
   
   Dbk_chkch(dbk);

   if(Dbk_getflag(dbk, DBK_F_EDITED) == FALSE && dbk->keyo == off && dbk->keyl == len) {
      for(; recc; recc >>= 1) {
         nbase = base + (recc >> 1) * recl;
         cmp = memcmp(key, nbase, len);
         if(cmp == 0)
            return nbase - off;
         if(cmp > 0) {
            base = nbase + recl;
            recc--;
         }
      }
   } else {   
      while(recc--) {
         if(memcmp(key, base, len) == 0)
            return base - off;    
         base += recl;
      }
   }
   return NULL;
}

int Dbk_recno(Dbk dbk, char *key) {
   char *orec = Dbk_get(dbk, key);
   return orec ? (int) (orec - dbk->datap) / dbk->recl : -1;
}

int Dbk_del(Dbk dbk, char *key) {
   char *mod  = "Dbk_del";
   char *orec =  Dbk_get(dbk, key);
   
   if(orec) {
      return Dbk_free(dbk, orec);
   } else {
      Dbk_errproc(dbk, mod, "[1] Key not found.");
   }
   return -1;
}

int Dbk_delrec(Dbk dbk, int no) {
   return Dbk_free(dbk, (char *)dbk->datap + (dbk->recl * no));
}

int Dbk_new(Dbk dbk, char *rec) {
   Dbk_chkch(dbk);
   dbk->datas += dbk->recl;
   dbk->datap  = (char *) Win_realloc(dbk->datap, dbk->datas);
   Dbk_set(dbk, (dbk->datap + (dbk->recl * (dbk->recc)++)), rec);
   Dbk_setflag(dbk, DBK_F_EDITED); 
   return 0;
} 

int Dbk_free(Dbk dbk, char *rec) {
   Dbk_chkch(dbk);
   memset(rec, 0xff, dbk->recl);
   Dbk_sortkey(dbk, 0, dbk->recl);
   dbk->datas -= dbk->recl;
   --(dbk->recc);
   return 0;
}

int Dbk_cmp(unsigned char *ptr1, unsigned char *ptr2) {
   unsigned int  len = Dbk_len;
   unsigned char *p1 = ptr1 + Dbk_off;
   unsigned char *p2 = ptr2 + Dbk_off;
   
   do {
      if(*p1++ != *p2++) return (*--p1 - *--p2) * Dbk_rev;   
   } while(--len);
    
   return 0;
}

void Dbk_sortparms(Dbk dbk, int o, int l, int r) {
   dbk->keyo = o;
   dbk->keyl = l;
   if(r) {
      Dbk_setflag(dbk, DBK_F_SORTR);
   } else {
      Dbk_unsetflag(dbk, DBK_F_SORTR);
   }   
   Dbk_setflag(dbk, DBK_F_EDITED);
}

long Dbk_mtime(char *filen) {
   struct stat s;
   s.st_mtime = 0;
   stat(filen, &s); 
   return s.st_mtime;
}

void Dbk_errproc(Dbk dbk, char *mod, char *err) {
   sprintf(Win_io, "%s[%s] -- %s", mod, dbk->filen ? dbk->filen : "<NoName>", strerror(errno));

   if(dbk->proc) {
      (*dbk->proc) (dbk, Win_io, err);
   } else {
      (void) fprintf(stderr,"%s: %s\n", Win_io, err);
   } 
}

void Dbk_chkch(Dbk dbk) {
   if(dbk->chpid) {
      waitpid(dbk->chpid, 0, 0);
      dbk->mtime = Dbk_mtime(dbk->filen);
      dbk->chpid = 0;
      Dbk_unsetflag(dbk, DBK_F_RDONLY);
   }
}

void Dbk_delobsolete(Dbk dbk) {
   struct dirent  *dirs;
   DIR            *dh;
   int             fcnt = 0;
   long            acttime = time(0);
         
   char *fname = strrchr(dbk->filen, '/');
   memset(Win_io, '\0', 1024);
 
   if(fname) {
      strncpy(Win_io, dbk->filen, fname - dbk->filen);
      ++fname;
   } else {
      strcpy(Win_io, "/");
      fname = dbk->filen;
   }
                     
   dh = opendir(Win_io);
         
   if(dh) {
      int len = strlen(fname);
            
      while((dirs = readdir(dh))) {
         if(memcmp(dirs->d_name, fname, len) == 0) {
            if(strlen(dirs->d_name) > len) {
               /* ignore lockfiles */
               if(dirs->d_name[len+1] == 'l') continue;
               /* aelter als 5 tage ? , oder mehr als 3 files ? */
               if(acttime - atol(dirs->d_name + len + 1) > 432000L || ++fcnt > 3) {
                  sprintf(Win_io, "%s%s", dbk->filen, dirs->d_name + len);
                  unlink(Win_io);
               }
               
            }
         }
      }
   }
}
#endif /* _Dbk_c */
