/* 
 * x-tlk Toolkit - Copyright (c) Holger Pfaff
 * http://pfaff.dnsalias.net
 *
 * This software maybe used for any purpose provided the
 * above copyright notice is retained. It is supplied as is.
 * No warranty expressed or implied - Use at your own risk.
 */

#ifndef _Dsp_c
#define _Dsp_c
#define _Tlk_

#include <tlk.h>

void Dsp_open(int *argc, char **argv, char *prog) {
   int             i;
   XrmValue        value;

   if(prog) Dsp_prog = prog ;
   Dsp_default_scale = DSP_SYSTEM_SCALE;

   for( i = 1; i < *argc; ++i) {
      if(!strcmp(argv[i],"-Wi")     ||
         !strcmp(argv[i],"-iconic")   )
         Win_iconic = TRUE;
      else
      if((!strcmp(argv[i],"-Wr")         ||
          !strcmp(argv[i],"-display")  ) &&
         *argc > i + 1 )
         Dsp_name = argv[i + 1];
      else   
      if((!strcmp(argv[i],"-Wp")          ||
          !strcmp(argv[i],"-position")  ) &&
         *argc > i + 2 )
         Win_x = atoi(argv[i + 1]),
         Win_y = atoi(argv[i + 2]);
      else   
      if((!strcmp(argv[i],"-WP")               ||
          !strcmp(argv[i],"-icon_position")  ) &&
         *argc > i + 2 )
         Win_ix = atoi(argv[i + 1]),
         Win_iy = atoi(argv[i + 2]);
      else   
      if((!strcmp(argv[i],"-Ws")      ||
          !strcmp(argv[i],"-size")  ) &&
         *argc > i + 2 )
         Win_w = atoi(argv[i + 1]),
         Win_h = atoi(argv[i + 2]);
      else   
      if((!strcmp(argv[i],"-Wf")      ||
          !strcmp(argv[i],"-font")  ) &&
         *argc > i + 1 )
         Dsp_fn[DSP_FONT_CNT - 1] = argv[i + 1],
         Dsp_default_scale = DSP_FONT_CNT - 1;
      else   
      if( !strcmp(argv[i],"-Wds") ||
          !strcmp(argv[i],"-slow")  )
         Dsp_slow = TRUE;
      else
      if( !strcmp(argv[i],"-bg") ||
          !strcmp(argv[i],"-color")  )
         Dsp_colname = argv[i + 1];
      else
      if((!strcmp(argv[i],"-Wx")       ||
          !strcmp(argv[i],"-scale")  ) &&
         *argc > i + 1 ) {
         Dsp_default_scale = atoi(argv[i + 1]);
         if(Dsp_default_scale >= DSP_FONT_CNT - 1 ||
            Dsp_default_scale <  0)
            Dsp_default_scale = DSP_SYSTEM_SCALE;
      }
   }
   
   XrmInitialize();
   Dsp = XOpenDisplay(Dsp_name);
   if(Dsp == 0) {
      fprintf(stderr, "%s: Unable to open display '%s'.\n", Dsp_prog, XDisplayName(Dsp_name));           
      Win_exit();
   }
   
   if(XResourceManagerString(Dsp)) Dsp_DB = XrmGetStringDatabase(XResourceManagerString(Dsp));
      
   Dsp_is_ow  = strncmp(ServerVendor(Dsp), "X11/NeWS", 8) ? FALSE : TRUE;
   Dsp_screen = DefaultScreen(Dsp);
   Dsp_depth  = DisplayPlanes(Dsp, Dsp_screen);
   Dsp_white  = WhitePixel(Dsp, Dsp_screen);
   Dsp_black  = BlackPixel(Dsp, Dsp_screen);
   Dsp_root   = RootWindow(Dsp, Dsp_screen);
   Dsp_color  = (Dsp_depth >= 8 && Dsp_slow == FALSE) ? TRUE : FALSE;
   Txt_color  = Dsp_color;

   for(i = 0; i < DSP_FONT_CNT; ++i) {
      Dsp_ft[i] = 0;
      Dsp_rw[i] = 0;
   }
         
   XSetErrorHandler(Win_error);
   
   Win_init();
   Txt_init();

   if(XrmGetResource(Dsp_DB, "*keyboard", "*", (char **)Win_uml, &value)) {
      Dsp_kbdtype = atoi(value.addr);
   }
}   

void Dsp_create_gcs() {
   XColor          col25, col50, col75;
   XGCValues       gcv;
   Colormap        cmap = DefaultColormap(Dsp, Dsp_screen);
   int             retc;
   char           *errmsg = " Can't allocate color ";

   /* Create GC' s */
   
   gcv.fill_style = FillTiled;
            

   if(Dsp_color) {
      /* Some versions of XAllocNamedColor seem to have problems
       * decoding numerical color definitions - so lets doit our-
       * self: #RRGGBB
       */
      if(Dsp_colname[0] == '#' && strlen(Dsp_colname) > 6) {
         Win_uml[2] = '\0';
         memcpy(Win_uml, &Dsp_colname[1], 2);
         col50.red   = (unsigned short) strtol(Win_uml, 0, 16) * 255;
         memcpy(Win_uml, &Dsp_colname[3], 2);
         col50.green = (unsigned short) strtol(Win_uml, 0, 16) * 255;
         memcpy(Win_uml, &Dsp_colname[5], 2);
         col50.blue  = (unsigned short) strtol(Win_uml, 0, 16) * 255;
         
         retc = XAllocColor(Dsp, cmap, &col50);
      } else {   
         retc = XAllocNamedColor(Dsp, cmap, Dsp_colname, &col50, &col25);
      }
      
      if(retc == 0) {
         fprintf(stderr, "%s:%s'%s'. Trying default.\n", Dsp_prog, errmsg, Dsp_colname);
         retc = XAllocNamedColor(Dsp, cmap, Dsp_colname=DSP_DEFAULT_COLOR, &col50, &col25);
      }
         
      /* we need a darker and lighter version of this color */

      col25.red   = col50.red   * DSP_COLORFACTOR > 0xffff ? 0xffff : col50.red   * DSP_COLORFACTOR;
      col25.green = col50.green * DSP_COLORFACTOR > 0xffff ? 0xffff : col50.green * DSP_COLORFACTOR;
      col25.blue  = col50.blue  * DSP_COLORFACTOR > 0xffff ? 0xffff : col50.blue  * DSP_COLORFACTOR;
      
      col75.red   = col50.red   / DSP_COLORFACTOR;
      col75.green = col50.green / DSP_COLORFACTOR;
      col75.blue  = col50.blue  / DSP_COLORFACTOR;
      
      retc += XAllocColor(Dsp, cmap, &col25);
      retc += XAllocColor(Dsp, cmap, &col75);
         
      Dsp_grey25 = col25.pixel;
      Dsp_grey50 = col50.pixel;
      Dsp_grey75 = col75.pixel;
      
      if(retc > 2) {
         gcv.fill_style = FillSolid;
      } else {
         fprintf(stderr, "%s:%s'%s'. Using b&w.\n", Dsp_prog, errmsg, Dsp_colname);
         Txt_color = Dsp_color = FALSE;
      }   
   } 
   
   gcv.graphics_exposures = FALSE;
   
   gcv.foreground = Dsp_white;
   gcv.background = Dsp_white;
   Dspgc_clr = XCreateGC(Dsp, Dsp_root, GCGraphicsExposures|GCForeground|GCBackground, &gcv);
   
   if(Txt_color) {            
      gcv.foreground = Dsp_black;  
      gcv.background = Dsp_grey75;
      Dspgc_inv = XCreateGC(Dsp, Dsp_root, GCGraphicsExposures|GCForeground|GCBackground, &gcv);
   } else {
      gcv.background = Dsp_black;
      gcv.function   = GXinvert;
      gcv.plane_mask = Dsp_black^Dsp_white;
      Dspgc_inv = XCreateGC(Dsp, Dsp_root, GCGraphicsExposures|GCFunction|GCForeground|GCBackground, &gcv);
   }
                                 
   gcv.foreground = Dsp_black;
   gcv.background = Txt_color ? Dsp_grey50 : Dsp_white;             
   Dspgc = XCreateGC(Dsp, Dsp_root, GCGraphicsExposures|GCForeground|GCBackground, &gcv);
               
   gcv.line_width = 2;     
   Dspgc_dbl = XCreateGC(Dsp, Dsp_root, GCGraphicsExposures|GCForeground|GCBackground|GCLineWidth, &gcv);
                                   
   if(Dsp_color) gcv.foreground = Dsp_grey25;
   gcv.tile = XCreatePixmapFromBitmapData(Dsp, Dsp_root, (char*)Dsp_gray25b, 2, 2, Dsp_black, Dsp_white, Dsp_depth);
   Dspgc_25 = XCreateGC(Dsp, Dsp_root, GCGraphicsExposures|GCTile|GCFillStyle|GCForeground, &gcv);
               
   if(Dsp_color) gcv.foreground = Dsp_grey50;
   gcv.tile = XCreatePixmapFromBitmapData(Dsp, Dsp_root, (char*)Dsp_gray50b, 2, 2, Dsp_black, Dsp_white, Dsp_depth );
   Dspgc_50 = XCreateGC(Dsp, Dsp_root, GCGraphicsExposures|GCTile|GCFillStyle|GCForeground, &gcv);
               
   if(Dsp_color) gcv.foreground = Dsp_grey75;
   gcv.tile = XCreatePixmapFromBitmapData(Dsp, Dsp_root, (char*)Dsp_gray75b, 2, 2, Dsp_black, Dsp_white, Dsp_depth );
   Dspgc_75 = XCreateGC(Dsp, Dsp_root, GCGraphicsExposures|GCTile|GCFillStyle|GCForeground, &gcv);
                              
   Dspcur[(int)DSPCUR_MNU] = XCreateFontCursor(Dsp, (unsigned int) XC_sb_right_arrow);
   Dspcur[(int)DSPCUR_ALR] = XCreateFontCursor(Dsp, (unsigned int) XC_question_arrow);
   Dspcur[(int)DSPCUR_NRM] = XCreateFontCursor(Dsp, (unsigned int) XC_top_left_arrow);
   Dspcur[(int)DSPCUR_BSY] = XCreateFontCursor(Dsp, (unsigned int) XC_watch);
   Dspcur[(int)DSPCUR_SCB] = XCreateFontCursor(Dsp, (unsigned int) XC_sb_v_double_arrow);
   Dspcur[(int)DSPCUR_HND] = XCreateFontCursor(Dsp, (unsigned int) XC_hand2);
   
   But_init();                           
}

int Dsp_load_font(int scale) {
   XFontStruct      *fs;

   /* Font already loaded? -> return TRUE */
   
   if(Dsp_ft[scale])
      return TRUE;
           
   if(Dspgc == 0)
      Dsp_create_gcs();
            
   /* Font not avail? -> return FALSE */
      
   if(Dsp_rw[scale] == DSP_FONT_NOT_AVAIL)
      return FALSE;      
      
   /* load Font */
   
   fs = XLoadQueryFont(Dsp, Dsp_fn[scale]);

   if(fs) {   
      Dsp_ft[scale] = fs->fid;
      Dsp_fd[scale] = fs->descent;
      Dsp_rw[scale] = fs->max_bounds.width;
      Dsp_rh[scale] = fs->ascent + fs->descent + 3;
      Dsp_ul[scale] = fs->max_char_or_byte2 >= XK_udiaeresis ? TRUE : FALSE;
                             
      /* check if fixed-width font */
                           
      if(fs->max_bounds.width - fs->min_bounds.width) {    
         fprintf(stderr, "%s: Font '%s' not fixed-width.\n", Dsp_prog, Dsp_fn[scale]);
         XFreeFont(Dsp, fs);             
         Dsp_ft[scale] = 0; 
      }
   }
        
   if(Dsp_ft[scale] == 0) {   
      fprintf(stderr, "%s: Font '%s' [%d] not found.\n", Dsp_prog, Dsp_fn[scale], scale); 
      if(scale == DSP_FONT_CNT - 1) Dsp_default_scale = DSP_SYSTEM_SCALE;
      Dsp_rw[scale] = DSP_FONT_NOT_AVAIL;
      return FALSE;
   } else {
      XSetFont(Dsp, Dspgc,     Dsp_ft[scale]);
      XSetFont(Dsp, Dspgc_25,  Dsp_ft[scale]);
      XSetFont(Dsp, Dspgc_50,  Dsp_ft[scale]);
      XSetFont(Dsp, Dspgc_75,  Dsp_ft[scale]);
      XSetFont(Dsp, Dspgc_inv, Dsp_ft[scale]);
      XSetFont(Dsp, Dspgc_clr, Dsp_ft[scale]);
      return TRUE;
   }
}

int Dsp_check_scale(int scale) {
   if(scale == DSP_DEFAULT_SCALE) scale  = Dsp_default_scale;
   if(scale >= DSP_FONT_CNT) scale  = DSP_FONT_CNT - 1;
   if(scale  < 0) scale  = 0;

   for(; scale < DSP_FONT_CNT; ++scale) if(Dsp_load_font(scale)) return scale;
   for(--scale; scale >= 0; --scale) if(Dsp_load_font(scale)) return scale;
   fprintf(stderr, "%s: No fonts found.\n" , Dsp_prog);
    
   Win_exit();
   return 0;
}

#endif /* Dsp_c */
