/* 
 * x-tlk Toolkit - Copyright (c) Holger Pfaff
 * http://pfaff.dnsalias.net
 *
 * This software maybe used for any purpose provided the
 * above copyright notice is retained. It is supplied as is.
 * No warranty expressed or implied - Use at your own risk.
 */

#ifndef _Lst_c
#define _Lst_c
#define _Tlk_

#include <tlk.h>

Lst Lst_create(Win win, int row, int col, int rows, int cols, void (*proc) (), char *label) {
   Lst lst;
   
   if(rows < 4) {
     fprintf(stderr, "Lst_create: rows set to 4");  
     rows = 4;
   }
   
   if(Lst_decoration) {
      Win_fb(win, Dspgc_50, row - 1, col - 1, rows + 2, cols + 9, 
             -Dsp_rw[win->scale]/2, Dsp_rh[win->scale]/2);
      Win_db(win, Dspgc,    row - 1, col - 1, rows + 2, cols + 9, 
             -Dsp_rw[win->scale]/2, Dsp_rh[win->scale]/2);
   }
   
   if(label) {
      Win_bs(win, Dspgc, row, col+5, Dsp_rw[win->scale]/2, -3, label);
   }

   if(Dsp_color) {
      Win_3d(win->pixmap, Dspgc_75, Dspgc_25, Dspgc_clr, 
             (col+5)  * Dsp_rw[win->scale] - 1, 
             (row+1)  * Dsp_rh[win->scale] - 1, 
             (cols+1) * Dsp_rw[win->scale] + 1, 
              rows    * Dsp_rh[win->scale] + 1);
              
      Win_3d(win->pixmap, Dspgc_75, Dspgc_25, Dspgc_clr, 
             (col+5)  * Dsp_rw[win->scale], 
             (row+1)  * Dsp_rh[win->scale], 
             (cols+1) * Dsp_rw[win->scale] - 1, 
              rows    * Dsp_rh[win->scale] - 1);
   } else {
      Win_db(win, Dspgc, row + 1, col + 5, rows, cols + 1, 0, +1);
      Win_dl(win, Dspgc, row + 1, col + 5, row + 1, col + cols + 6, 0, 0);
   }
                       
   lst = Win_create(win, WIN_LST_WIN, row + 1, col + 5, rows, cols + 1, win->scale, 0, 0, proc);
   lst->cols  = cols;
   lst->x     = -1; /* selected item */
   lst->y     =  0; /* scroll offset */
   
   lst->cnt   = (int) Mnu_create(lst_mnu, 0, DSP_DEFAULT_SCALE, "List");
   
   lst->data  = (char*) Dbk_create(cols, 0, cols, 0);  
   lst->label = (char*) Scb_create(win, row+1, col, rows, lst, Lst_scb);
               
   return lst;
}

void Lst_mnu_proc(Mnu m, char* l, int i) {   
   switch(i) {
    case -1:  
       break;
    case  0:  /* sort */
    case  1:  /* sort rev */
       Lst_sortparms(Lst_cur, 0, Lst_cur->cols, i);
       Lst_sort(Lst_cur);
      
    case  3:
       Lst_cur->x = -1;
          
    case  2:  /* redisp */
       Lst_dsp(Lst_cur);
       break;
       
    default:
       if(Lst_cur->proc) (void)(Lst_cur->proc)(Lst_cur, l, i);
       break;
   };
}


void Lst_dsp(Lst lst) {
   int i;
   char f[20]; 

   sprintf(f, "%%-%d.%ds", lst->cols, lst->cols);
   
   if(lst->y + lst->rows > Lst_cnt(lst)) {
      lst->y = Lst_cnt(lst) - lst->rows;
      if(lst->y < 0) lst->y = 0;
   }
   
   Tlk_set(Tlk_get(lst->cnt, TLK_GET_MNI, 3), lst->x > -1 ? TLK_ACTIVATE : TLK_DEACTIVATE);

   for(i = 0; (i+lst->y) < Lst_cnt(lst) && i < lst->rows; ++i) {
      sprintf(Win_buf, f, Lst_get(lst, i + lst->y));
      Win_ft(lst, (i + lst->y == lst->x) ? Dspgc_inv : Dspgc, 
             i, 0, Dsp_rw[lst->scale]/2, 0, Win_buf, lst->cols);
   }
   
   Scb_set(lst->label, Lst_cnt(lst), lst->rows, lst->y);
}

void Lst_updn(Lst lst, int updn) {
   int val = lst->y + updn;
   int max = Lst_cnt(lst) - lst->rows;
        
   if(val < 0) {
      if(lst->y > 0) {
         val = 0;
      } else {
         return;
      }
   }
         
   if(val > max) {
      if(lst->y < max) {
         val = max;
      } else {
         return;
      }
   }
         
   if(updn) {
      lst->y = val;
      Lst_dsp(lst);
   }  
}

void Lst_reset(Lst lst) {
   Dbk dbk = (Dbk) lst->data;
   int keyo = dbk->keyo;
   int keyl = dbk->keyl;
   int sortr= Dbk_getflag(dbk, DBK_F_SORTR);
   Dbk_destroy(dbk);
   lst->data  = (char*) Dbk_create(lst->cols, keyo, keyl, sortr);  
   XClearWindow(Dsp, lst->window);
}


void Lst_scb(Scb scb, int cnt) {
   Lst_updn((Lst) scb->label, cnt);  
}

void Lst_key_press(Lst lst, KeySym keysym, char* str, int len) {
   int  updn = 0;
   
   switch(keysym) {
      case XK_Begin:
      case XK_Home:
      case XK_R7:
               updn -= lst->y;
               break;
      case XK_End:
      case XK_R13:
               updn  = Lst_cnt(lst) - lst->y;
               break;
      case XK_Up:
      case XK_R8:
               updn = 1;
               break;
      case XK_Down:
      case XK_R14:
               updn = -1;
               break;
      case XK_Next:
      case XK_R15:
               updn =  lst->rows;
               break;
      case XK_Prior:
      case XK_R9:
               updn = -lst->rows;
               break;
      case XK_Return:
      case XK_KP_Enter:
               if(lst->x > -1 && lst->proc) {
                 (void) (lst->proc) (lst, Dbk_getrec(lst->data, lst->x), LST_NTFY_DOUBLE);
               }
      case XK_Linefeed:
      case XK_Tab:
               Lst_charcnt = 0;
               break;
      default:
               if(len && (Lst_charcnt + len) < LST_CHARCNT) {
                  if(Txt_is_alpha(str, len)) {
                     memcpy(&Lst_char[Lst_charcnt], str, len);
                     Lst_charcnt += len;
                  }
                  if(keysym == XK_BackSpace || 
                     keysym == XK_Delete) {
                      if(Lst_charcnt) --Lst_charcnt;
                  }
                  if(Lst_charcnt) {
                     Lst_sortparms(lst, 0, Lst_charcnt, 0);
                     Lst_hilite(lst, Dbk_recno(lst->data, Lst_char));
                  }
               }
               break;
   }
   Lst_updn(lst, updn);
}

void Lst_button_press(Lst lst, XButtonEvent* xb) {
   if(xb->button == Button3) {
      But_xbutton_pressed = 1;
      Lst_cur = lst;
      Mnu_map(lst->cnt, Lst_mnu_proc);
      return;
   }

   if(xb->button == Button2) return;

   xb->y =  xb->y / Dsp_rh[lst->scale] + lst->y; /* actual row */
      
   /* doubleclick ? */
   
   if(lst->x == xb->y && xb->time - Lst_time < 500) {
      if(lst->proc) {
        (void)(lst->proc)(lst, Dbk_getrec(lst->data, xb->y), LST_NTFY_DOUBLE);
      }
   }
   
   Lst_hilite(lst, xb->y);
   
   while(XQueryPointer(Dsp, lst->window,
                      &xb->root,      &xb->window,
                      &xb->x_root,    &xb->y_root,
                      &xb->x, &xb->y, &xb->state)) {
      if((xb->state & Button1Mask) == 0) break;   /* Button losgelassen */  
            
      if(xb->y < 0) {
         Lst_updn(lst, -1);
         Win_sleep(50);
         continue;
      }

      xb->y /= Dsp_rh[lst->scale]; /* actual row */
   
      if(xb->y >= lst->rows) {
         Lst_updn(lst, 1);
         Win_sleep(50);
         continue;
      }
      
      Lst_hilite(lst, xb->y + lst->y);      
   }
   Lst_time = xb->time;
}

void Lst_hilite(Lst lst, int x) {
   Txt_unset_cursor();
   
   /* Is it a different row */
   
   if(x >= 0 && x < Lst_cnt(lst) && lst->x - x) {
      
      if(lst->x >= 0 && lst->x < Lst_cnt(lst)) {
         lst->x = -1;
         if(lst->proc) (void)(lst->proc)(lst, Lst_get(lst, lst->x), LST_NTFY_DESELECT);
         Tlk_set(Tlk_get(lst->cnt, TLK_GET_MNI, 3), TLK_DEACTIVATE);
         Lst_dsp(lst);
      }
      
      if(x < Lst_cnt(lst)) { 
         lst->x = x;
         if(lst->proc) (void)(lst->proc)(lst, Lst_get(lst, lst->x), LST_NTFY_SELECT);
         Tlk_set(Tlk_get(lst->cnt, TLK_GET_MNI, 3), TLK_ACTIVATE);
         Lst_dsp(lst);
         
         /* Make selected item visible */
         
         if(lst->x - lst->y < 0) Lst_updn(lst, -(lst->y - lst->x + lst->rows/2));
         if(lst->x - lst->y >= lst->rows) Lst_updn(lst, -(lst->y - lst->x + lst->rows - lst->rows/2 - 1));
      }
   }
}

void Lst_sortparms(Lst lst, int o, int l, int r) {
   Dbk_sortparms(lst->data, o, l, r);
    
   Tlk_set(Tlk_get(lst->cnt, TLK_GET_MNI, 0), l ? TLK_ACTIVATE : TLK_DEACTIVATE);
   Tlk_set(Tlk_get(lst->cnt, TLK_GET_MNI, 1), l ? TLK_ACTIVATE : TLK_DEACTIVATE);
}   

#endif            /* _Lst_c */
