/* 
 * x-tlk Toolkit - Copyright (c) Holger Pfaff
 * http://pfaff.dnsalias.net
 *
 * This software maybe used for any purpose provided the
 * above copyright notice is retained. It is supplied as is.
 * No warranty expressed or implied - Use at your own risk.
 */

#ifndef _Mnu_c
#define _Mnu_c
#define _Tlk_ 

#include <tlk.h>

Mnu Mnu_create(char **items, int rows, int scale, char *headline) {
   Mnu      mnu;
   int      item_len = 0;
   int      item_cnt = 0;

   if(items) {
      while(items[item_cnt]) {     
         if(strlen(items[item_cnt]) > item_len) item_len = strlen(items[item_cnt]);
         ++item_cnt;
      }
   } else {
      item_len = rows;
   }
            
   if(headline && strlen(headline) > item_len + 2) item_len = strlen(headline);
         
   item_len += 3;
      
   mnu = Win_create(NULL, WIN_MNB_WIN, 0, 0, 1, item_len, scale, NULL, headline, NULL);

   mnu->parent = mnu;
   mnu->rows   = items == 0 ? 0 : rows;
   mnu->data   = NULL;
   mnu->cnt    = item_cnt = 0;

   while(items && items[item_cnt]) Mnu_append_item(mnu, items[item_cnt++], NULL);
      
   return mnu;
}

void Mnu_append_item(Mnu mnu, char *item, Mnu smnu) {
   Mnu   ptr;
   int   rows, cols;
   int   r,c,l=item?strlen(item)+1:1;
   int   item_cnt;
   int   item_len;
   int   hr = 0;

   /* Make sure menu is wide enough */
   item_len = !mnu->data ? mnu->w / Dsp_rw[mnu->scale] : ((Mnu) mnu->data)->cols;
   if(item_len < l) item_len=l;
   item_len+= smnu ? 2 : 0;

   ptr      = mnu->parent;
   item_cnt = mnu->cnt;

   ptr->data = (caddr_t) Win_create(mnu, WIN_MNI_WIN, 0, 0, 1, item_len, mnu->scale, 0, item, 0 );
   if(l == 1) Tlk_set(ptr->data, TLK_DEACTIVATE);
   ptr         = (Mnu) ptr->data;
   ptr->cnt    = item_cnt++;
   mnu->cnt    = item_cnt;
   mnu->parent = ptr;
   
   ptr->middle = (int *) smnu;
   if(smnu) Win_setflag(ptr, WIN_F_HAS_SUBMENU);
   rows  = mnu->rows == 0 ? item_cnt : mnu->rows;

   cols  = item_cnt / rows; 
   cols +=(item_cnt % rows) == 0 ? 0 : 1;

   /* menu should not become larger than screen */
   if(rows  * Dsp_rh[mnu->scale] > DisplayHeight(Dsp, Dsp_screen)) {
      rows  = DisplayHeight(Dsp, Dsp_screen) / Dsp_rh[mnu->scale];
      cols  = item_cnt   / --rows; 
      cols += !(item_cnt % rows) ? 0 : 1;
      rows  = item_cnt   / cols;
      rows += !(item_cnt % cols) ? 0 : 1;
   }
      
   /* add one row if we have a label */
   if(mnu->label) hr = 1, ++rows;
   
   mnu->cols = cols * item_len;
   mnu->w    = MAX((mnu->cols * Dsp_rw[mnu->scale]) + (Dsp_color ? 5 : 1), mnu->w);
   mnu->h    = (rows * Dsp_rh[mnu->scale]) + (Dsp_color ? 5 : 1);
   
   ptr = (Mnu) mnu->data;

   for(c=0; c < cols && ptr; ++c)  {
      for(r=hr; r < rows && ptr; ++r) {
         XMoveWindow(Dsp, ptr->window, 
                     ptr->x = c * Dsp_rw[mnu->scale] * item_len + (Dsp_color ? 2 : 0), 
                     ptr->y = r * Dsp_rh[mnu->scale] + (Dsp_color ? 2 : 0));
         ptr = (Mnu) ptr->data;
      }     
   }
}

void Mnu_map(Mnu mnu, void (*proc)()) {
   mnu->x = Win_lastx + mnu->w > DisplayWidth(Dsp, Dsp_screen) ? 
            DisplayWidth(Dsp, Dsp_screen) - mnu->w - 7 : Win_lastx + 3;
               
   mnu->y = Win_lasty + mnu->h > DisplayHeight(Dsp, Dsp_screen) ?
            DisplayHeight(Dsp, Dsp_screen) - mnu->h - 7 : Win_lasty - Dsp_rh[mnu->scale];

   XMoveResizeWindow(Dsp, mnu->window, mnu->x + 1, mnu->y + 1, mnu->w, mnu->h);
          
   if(Mnu_act < MNU_MAX) {
    ++Mnu_act;
   } else {
      fprintf(stderr, "%s: Too many menu-levels\n", Dsp_prog);
   }

   mnu->proc         = proc;
   Mnu_item_selected = FALSE;
   Mnu_stay_up       = TRUE;
   Mnus[Mnu_act]     = mnu;

   XMapRaised(Dsp, mnu->window);
   
   if(Mnu_act == 0) {
      XGrabPointer(Dsp, mnu->window, TRUE,
           ButtonPressMask | ButtonReleaseMask |
           EnterWindowMask | LeaveWindowMask,
           GrabModeAsync, GrabModeAsync, None,
           Dspcur[(int)DSPCUR_MNU], CurrentTime);

      XGrabKeyboard(Dsp, mnu->window, TRUE,
           GrabModeAsync, GrabModeAsync, CurrentTime);
   }
}

void Mnu_unmap(XButtonEvent *xbutton) {
   if(Mnu_stay_up &&
     (abs(Win_lastx - (xbutton ? xbutton->x_root : 0)) +
      abs(Win_lasty - (xbutton ? xbutton->y_root : 0))) < 2)
      goto Mnu_umanp_exit;

   if(Mnu_act != -1) {
      XUngrabPointer(Dsp, CurrentTime);
      XUngrabKeyboard(Dsp, CurrentTime);
      XFlush(Dsp);

      if(But_active) {
         But_draw(But_active);
         But_active = NULL;
      }
      
      do {
         XUnmapWindow(Dsp, Mnus[Mnu_act]->window);
      } while(Mnu_act--);

      if(Mnu_item_selected) {
         if(Win_c->parent) {
           if(Win_c->parent->proc) {
             (void)(*Win_c->parent->proc)(Win_c->parent, Win_c->label, Win_c->cnt);
           }
         }
      } else {
         if(Mnus[Mnu_act + 1]->proc) (void)(*Mnus[Mnu_act + 1]->proc)(Mnus[Mnu_act + 1], Mnus[Mnu_act + 1]->label, -1);
      }

      if(Txt_active) {
         Win_ft(Txt_active->win, Dspgc, Txt_active->row, Txt_active->col, 0, 0, Txt_active->data, Txt_active->len);
      }
   }
   
Mnu_umanp_exit:
   if(Mnu_act == 0) Mnu_stay_up = FALSE;
}

#endif /* _Mnu_c */
