/* 
 * x-tlk Toolkit - Copyright (c) Holger Pfaff
 * http://pfaff.dnsalias.net
 *
 * This software maybe used for any purpose provided the
 * above copyright notice is retained. It is supplied as is.
 * No warranty expressed or implied - Use at your own risk.
 */

#ifndef _Scb_c
#define _Scb_c
#define _Tlk_

#include <tlk.h>

Scb Scb_create(Win win, int row, int col, int rows, char *data, void (*proc)()) {
   But butup;
   But butdn;
   Scb scb;
   
   if(rows <  4) return 0;
   if(proc == 0) proc = Scb_dsp;

   scb = Win_create(win, WIN_BUT_WIN, row + 2, col, rows - 3, 4, win->scale, NULL, data, proc);          

   butup = But_create(win, row,            col, "/\\", 0, Scb_but);
   butdn = But_create(win, row + rows - 1, col, "\\/", 0, Scb_but);

   butup->type = butdn->type = WIN_RBU_WIN;
   butup->data = butdn->data = (char*) scb;
   butup->cnt  = -1;
   butdn->cnt  =  1; 

   XDefineCursor(Dsp, scb->window, Dspcur[(int)DSPCUR_SCB]);
   XMoveWindow(Dsp, scb->window, 
                    scb->x, scb->y = scb->y - (Dsp_rh[scb->scale]/2) - (Dsp_color ? 3 : 0) + .5);
          
   scb->type = WIN_SCB_WIN;
   scb->x    = 0;   /* units displayed */
   scb->y    = 0;   /* position */
   scb->cnt  = 0;   /* units total */
   
   scb->h   -= 6;
   scb->w   -= (Dsp_color ? 6 : 7);
   
   return scb;
}

void Scb_but(But but) {
   (void)(((Scb)but->data)->proc)((Scb)(but->data), but->cnt);
}

void Scb_calc(Scb scb) {   
   if(scb->cnt) {
      float f = (float) scb->x / (float) scb->cnt;
      if(f > 1) f = 1;
      Scb_s2h  = f * scb->h;
      if(Scb_s2h<8) Scb_s2h = scb->h > 8 ? 8 : scb->h;
      f = (float) scb->y / (float) scb->cnt;
      Scb_s2y1 = f * scb->h;
      if(Scb_s2y1 + Scb_s2h > scb->h) Scb_s2y1 = scb->h - Scb_s2h;
      Scb_s2y2 = Scb_s2y1 + Scb_s2h;
   } else {
      Scb_s2h  = scb->h;
      Scb_s2y1 = 0;
      Scb_s2y2 = scb->h;
   }
   Scb_s2y1 += 3;
}

void Scb_bar(Scb scb, GC gc) {
   XFillRectangle(Dsp, scb->window, gc, 3, Scb_s2y1, scb->w, Scb_s2h);   
}

void Scb_dsp(Scb scb) {
   if(Dsp_color)
      But_3d(scb, Dspgc_25, Dspgc_75, Dspgc_50);
   else
      XClearWindow(Dsp, scb->window);
      
   Scb_calc(scb);
   Scb_bar(scb, Dspgc_75);
}

void Scb_set(Scb scb, int nel, int vel, int pos) {
   scb->cnt = nel;  /* No.  of elements */
   scb->x   = vel;  /* Visible elements */
   scb->y   = pos;  /* current position */
   Scb_dsp(scb);
}
   
void Scb_button_press(Scb scb, XButtonEvent *xb) {      
   if(xb->y < Scb_s2y1) {
      (void)(scb->proc)(scb, -scb->rows);
      return;
   }
   if(xb->y > Scb_s2y2) {
      (void)(scb->proc)(scb,  scb->rows);
      return;
   }   
   Scb_calc(scb);
   xb->type = xb->y;  /* Backup current pos in type */
   Scb_bar(scb, Dspgc);
   while(XQueryPointer(Dsp, scb->window,
                      &xb->root,      &xb->window,
                      &xb->x_root,    &xb->y_root,
                      &xb->x, &xb->y, &xb->state)) {
       if((xb->state & Button1Mask) == 0) break;   /* Button losgelassen */       
       if(xb->y - xb->type) {
          float f = (float) (xb->y - xb->type) / (float) scb->h;          
          (void)(scb->proc)(scb, (int)(f*scb->cnt));
          xb->type = xb->y;
          Scb_bar(scb, Dspgc);
       }
    }
    Scb_dsp(scb);        
}

#endif            /* _Scb_c */
