/* 
 * x-tlk Toolkit - Copyright (c) Holger Pfaff
 * http://pfaff.dnsalias.net
 *
 * This software maybe used for any purpose provided the
 * above copyright notice is retained. It is supplied as is.
 * No warranty expressed or implied - Use at your own risk.
 */

#ifndef _Tlk_c
#define _Tlk_c
#define _Tlk_

#include <tlk.h>

void Tlk_set(Win win, Tlk_set_t what, caddr_t arg1, caddr_t arg2) {
   Window              window;
   Txt                 txt = 0;
   XrmValue            value;

   if(win) {
      window = win->window;
      txt    = (Txt) win;
   }
   switch(what) {
     case TLK_SET_XY:
       XMoveWindow(Dsp, window, (int) arg1, (int) arg2);
       break;
     case TLK_SET_BORDER:
     case TLK_UNSET_BORDER:
       XSetWindowBorder(Dsp, window, 
       what == TLK_SET_BORDER ? BlackPixel(Dsp, Dsp_screen) : Dsp_color ? Dsp_grey25 : WhitePixel(Dsp, Dsp_screen));
     case TLK_SHOW:
     case TLK_OPEN:
       switch(win->type) {
         case WIN_POP_WIN:
         case WIN_SUB_WIN:
         case WIN_BUT_WIN:
         case WIN_ALB_WIN:
         case WIN_BAS_WIN:
         case WIN_ABU_WIN:
           XMapRaised(Dsp, window);
           break;
         case WIN_MNB_WIN:
           Mnu_map(win, (void (*) ()) arg1);
         default:
           break;
       }
       break;
     case TLK_UNSHOW:
       switch(win->type) {
         case WIN_POP_WIN:
         case WIN_SUB_WIN:
         case WIN_BUT_WIN:
         case WIN_ABU_WIN:
         case WIN_ALB_WIN:
           XUnmapWindow(Dsp, window);
         default:
           break;
       }
       break;
     case TLK_NORMALIZE:
       while((int) arg1-- && txt) {
         if(txt->flag == TRUE) {
            Win_ft(txt->win, Dspgc, txt->row, txt->col, 0, 0, txt->data, txt->len); txt->flag = FALSE;
         }
         txt = txt->next;
       }
       break;
     case TLK_INVERT:
       while((int) arg1-- && txt) {
         if(txt->flag == FALSE) {
            Win_ft(txt->win, Dspgc_inv, txt->row, txt->col, 0, 0, txt->data, txt->len); txt->flag = TRUE;
         }
         txt = txt->next;
       }
       break;
     case TLK_CLOSE:
       {
         XClientMessageEvent ev;
         Atom                atom;
         atom = XInternAtom(Dsp, "WM_CHANGE_STATE", False);
         if(atom == None) break;
         ev.type         = ClientMessage;
         ev.window       = window;
         ev.message_type = atom;
         ev.format       = 32;
         ev.data.l[0]    = IconicState;
         XSendEvent(Dsp, RootWindow(Dsp, Dsp_screen), False, SubstructureRedirectMask|SubstructureNotifyMask, (XEvent *) & ev);
       }
       break;
     case TLK_DEACTIVATE:
       switch(win->type) {
         case WIN_BAS_WIN:
         case WIN_POP_WIN:
         case WIN_SUB_WIN:
           Txt_unset_cursor();
           Txt_noedits = TRUE;
           Txt_active  = NULL;
           break;
         case WIN_BUT_WIN:
         case WIN_ABU_WIN:
         case WIN_MNI_WIN:
           Win_setflag(win, WIN_F_DEACTIVATED);
           XSelectInput(Dsp, window, (long) ExposureMask);
         default:
           break;
       }
       break;
     case TLK_SET_BUT_RPT:
       win->type = WIN_RBU_WIN;
       break;
     case TLK_ACTIVATE:
       switch(win->type) {
         case WIN_BUT_WIN:
         case WIN_ABU_WIN:
         case WIN_MNI_WIN:
           Win_unsetflag(win, WIN_F_DEACTIVATED);
           XSelectInput(Dsp, window, (long) (OwnerGrabButtonMask|EnterWindowMask|ButtonPressMask|LeaveWindowMask|ButtonReleaseMask|ExposureMask));
           break;
         case WIN_BAS_WIN:
         case WIN_POP_WIN:
         case WIN_SUB_WIN:
           Txt_noedits = FALSE;
         default:
           break;
       }
       break;
     case TLK_SET_LABEL:
       win->label = (char *) arg1;
       switch(win->type) {
         case WIN_BUT_WIN:
         case WIN_ABU_WIN:
         case WIN_MNI_WIN:
           But_draw(win);
         default:
           break;
       }
       break;
     case TLK_APPEND_MNI:
       Mnu_append_item(win, (char *) arg1, (Mnu) arg2);
       break;
     case TLK_SET_CURSOR:
       XDefineCursor(Dsp, window, Dspcur[(int) arg1]);
       break;
     case TLK_SET_MNU_ROWS:
       win->rows = (int) arg1;
       break;
     case TLK_SET_DEFAULT_SCALE:
       if(Dsp_default_scale == DSP_SYSTEM_SCALE) Dsp_default_scale = (int) arg1;
       break;
     case TLK_SET_DEFAULT_COLOR:
       if(arg1[0] == '$')  {
          if(Dsp_DB) {
             if(XrmGetResource(Dsp_DB, &arg1[1], "*", (char **) Win_uml, &value)) {
                strcpy(Win_buf, (char*) value.addr);
             } else {
                break;
             }
          } else {
             break;
          }
       } else {
          strcpy(Win_buf, (char*) arg1);
       }
       if(strcmp(Dsp_colname, DSP_DEFAULT_COLOR) == 0) Dsp_colname = Win_buf;
       break;
     case TLK_SET_DEFAULT_FONT:
       if(Dsp_default_scale == DSP_SYSTEM_SCALE) {
          strcpy(Win_io, (char *) arg1);
          Dsp_fn[DSP_FONT_CNT - 1] = Win_io;
          Dsp_default_scale = DSP_FONT_CNT - 1;
       }
       break;
     case TLK_SET_MIDDLE_PROC:
       Win_middle_proc = (void (*) ()) arg1;
       break;
     case TLK_SET_TXT:
       Txt_set_txt(txt, (char *) arg1);
       break;
     case TLK_SET_TXT_MODE:
       Txt_mode = (Txt_mode_t) arg1;
       break;
     case TLK_SET_TXT_STATUS:
       txt->status = (Txt_status_t) arg1;
       break;
     case TLK_REDISP_ALL:
     case TLK_SET_ALL_TXT:
       txt = txt ? txt : Txts;
       while(txt->next) {
          if(what == TLK_REDISP_ALL) {
             Win_ft(txt->win, Dspgc, txt->row, txt->col, 0, 0, txt->data, txt->len);
          } else {
             Txt_set_txt(txt, arg1 + (txt->data - txt->win->data));
          }
          txt = txt->next;
       }
       break;
     case TLK_SET_TXT_MISC_PROC:
       Txt_misc_proc = (void (*)()) arg1;
       break;
     case TLK_SET_TXT_EDIT_PROC:
       Txt_edited_proc = (void (*)()) arg1;
       break;
     case TLK_SET_TXT_EDIT_FLAG:
       Txt_edited = (int) arg1;
       break;
     case TLK_SET_TXT_CLEAR:
       Txt_clear = (int) arg1;
       break;
     case TLK_BELL:
       XBell(Dsp, 100);
       break;
   }
   XFlush(Dsp);
}

caddr_t Tlk_get(Win obj, Tlk_get_t what, caddr_t arg1) {
   Win ptr;
   Txt txt;
   char *chr;
   int i;

   if(obj == 0) return (caddr_t) "";

   switch(what) {
     case TLK_GET_MNI:
       ptr = (Win) obj->data;
       while(ptr) {
         if(ptr->cnt == (int) arg1)
            return (caddr_t) ptr;
         ptr = (Win) ptr->data;
       }
       return NULL;
     case TLK_GET_LABEL:
     case TLK_GET_LABEL_P:
       switch(obj->type) {
         case WIN_BUT_WIN:
         case WIN_MNI_WIN:
         case WIN_ABU_WIN:
           return (caddr_t) obj->label;
         case WIN_TXT_WIN:
           txt = (Txt) obj;
           memset(Win_io, '\0', WIN_IO_BUFSIZE);
           if(what == TLK_GET_LABEL_P) {
              chr = txt->data; i = 0;
              while(chr < txt->data + txt->len) {
                 if(*chr == '(' || *chr == ')') *(Win_io + i++) = '\\';
                 *(Win_io + i++) = *chr++;
              }
           } else {
              memcpy(Win_io, txt->data, txt->len);
           }
           chr = Win_io + txt->len - 1;
           while(chr >= Win_io) {
             if(*chr  == ' ')  {
                *chr = '\0';
             } else {
                break;
             }
             chr--;
           }
           return (caddr_t) Win_io;
         default:
           break;
       }
     case TLK_GET_VENDOR:
       return (caddr_t) ServerVendor(Dsp);
     default:
       break;
   }
   return 0;
}

#endif            /* _Tlk_c */
