/* 
 * x-tlk Toolkit - Copyright (c) Holger Pfaff
 * http://pfaff.dnsalias.net
 *
 * This software maybe used for any purpose provided the
 * above copyright notice is retained. It is supplied as is.
 * No warranty expressed or implied - Use at your own risk.
 */

#ifndef _Tlk_h
#define _Tlk_h

#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <X11/Xresource.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <dirent.h>
#include <string.h>
#include <malloc.h>
#include <signal.h>
#include <ctype.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>

#ifndef TRUE
#define TRUE   1
#endif

#ifndef FALSE
#define FALSE   0
#endif

#define MAX(a,b) (a > b ? a : b)


/* Window types for use with Win_create - use only first three 
 * others are for internal us only
 */
typedef enum {
   WIN_BAS_WIN,            /* BASE - Window  */
   WIN_POP_WIN,            /* POPUP - Window - no decoration */
   WIN_SUB_WIN,            /* SUB - Window - no decoration  */

   WIN_BUT_WIN,            /* BUTTON - Window       */
   WIN_MNB_WIN,            /* MENU_BASE - Window    */
   WIN_MNI_WIN,            /* MENU_ITEM - Window    */
   WIN_ALB_WIN,            /* ALARM_BASE - Window   */
   WIN_ABU_WIN,            /* ALARM_BUTTON - Window */
   WIN_TXT_WIN,            /* TEXT_ITEM - Window    */
   WIN_RBU_WIN,            /* REPEAT_BUTTON - Window  */
   WIN_LST_WIN,            /* LIST - Window  */
   WIN_SCB_WIN             /* Scrollbar - Window  */
} Win_t;

/* Return values from alarm windows */
typedef enum {
   WIN_ALARM_NO_BUT,         /* error occured */
   WIN_ALARM_BUT1,           /* button1 was selected */
   WIN_ALARM_BUT2,           /* button2 was selected */
   WIN_ALARM_BUT3            /* button3 was selected */
} Win_alarm_but_t;

/* possible txt field types */
typedef enum {
   TXT_RO_NUM,    /* read only - numerical field */
   TXT_RW_NUM,    /* read/write - numerical field */
   TXT_RO_ALPHA,  /* read only - general field */
   TXT_RW_ALPHA   /* read/write - general field */
} Txt_t;

/* possible cursor (mouse) types */
typedef enum {
   DSPCUR_NRM,   /* normal cursor */
   DSPCUR_MNU,   /* menu cursor */
   DSPCUR_ALR,   /* alarm cursor */
   DSPCUR_BSY,   /* busy cursor */
   DSPCUR_SCB,   /* scrollbar cursor */
   DSPCUR_HND    /* hand cursor */
} Dspcur_t;

/* txt type - internal */
typedef enum {
   TXT_TYPE_NUM,   /* numerical field */
   TXT_TYPE_ALPHA, /* general field */
   TXT_TYPE_PASSW  /* password field */
} Txt_type_t;

/* txt status - internal */
typedef enum {
   TXT_STATUS_RO,  /* read only */
   TXT_STATUS_RW   /* read/write */
} Txt_status_t;

/* txt mode - internal */
typedef enum {
   TXT_MODE_INS,   /* insert mode */
   TXT_MODE_OVR    /* overwrite mode */
} Txt_mode_t;

/* values send to list notify proc */
typedef enum {
   LST_NTFY_DOUBLE   = -1, /* double click occured */
   LST_NTFY_DESELECT = -2, /* deselection */
   LST_NTFY_SELECT   = -3  /* selection */
} Lst_mode_t;

#define WIN_F_DEACTIVATED      (1<<0)    /* Window deactivated */
#define WIN_F_HAS_SUBMENU      (1<<1)    /* Window has Submenu */
#define WIN_F_IS_MAPPED        (1<<2)    /* Window is mapped (visible) */

#define Win_getflag(win, f)    (((Win)win)->flags &   f)
#define Win_setflag(win, f)    (((Win)win)->flags |=  f)
#define Win_unsetflag(win, f)  (((Win)win)->flags &= !f)

#define Lst_sort(L)              Dbk_sort((Dbk)L->data)
#define Lst_append_mnu(L, l, sm) Mnu_append_item((Dbk)L->cnt, l, sm)
#define Lst_delete(L, i)         Dbk_delrec((Dbk)L->data, i)
#define Lst_get(L, i)            (char *) Dbk_getrec(L->data, i)
#define Lst_insert(L, s)         Dbk_put((Dbk)L->data, s, DBK_MULTIPLE)
#define Lst_replace(L, s)        Dbk_put((Dbk)L->data, s, DBK_REPLACE)
#define Lst_redisp(L)            XClearWindow(Dsp, L->window), Lst_dsp(L, 0)
#define Lst_cnt(L)               Dbk_cnt((Dbk)L->data)
#define Lst_sel(L)               L->x
#define Lst_getsel(L)            (char *) (L->x == -1 ? NULL : Lst_get(L, L->x))

typedef struct _Win_s {
   Win_t          type;     /* Window-Type                 */

   Window         window;   /* Window ID frm XCreateWindow */
   Pixmap         pixmap;   /* Backup Pixmap for window    */

   int            x;        /* x,y - Postion in Pixel      */
   int            y;

   int            w;        /* width, height in Pixel      */
   int            h;
  
   int            rows;     /* No. of rows                 */
   int            cols;     /* No. of cols                 */

   int            scale;    /* Scale (0 - 8)               */

   int            cnt;      /* count                       */
   
   int            flags;    /* div. flags                  */

   char          *data;     /* textdata                    */

   void         **text;     /* textfields                  */
   int           *middle;   /* middle mouse handle         */

   char          *label;    /* label                       */
   void         (*proc)();  /* callback func               */

   struct _Win_s *parent;   /* linked list of all wins     */
   struct _Win_s *next;
} Win_s;

typedef Win_s *  Win;
typedef Win_s *  But;
typedef Win_s *  Mnu;
typedef Win_s *  Mni;
typedef Win_s *  Lst;
typedef Win_s *  Scb;

typedef struct _Txt_s {
   Win_t          wtype;      /* Window-Type ( =WIN_TXT_WIN) */

   Win            win;        /* Parentwindow                */
    
   int            row, col; 
   int            len;     
   
   int            flag;   

   char          *data;       /* ptr to data                 */
 
   Txt_type_t     type;       /* Numeric or Alpha or Passwd  */
   Txt_status_t   status;     /* RO or RW                    */

   Mnu            mnu;        /* Assoc menu                  */
 
   void         (*proc)();    /* callback func               */

   struct _Txt_s *next;       /* linked list of all txt flds */
   struct _Txt_s *prev;
} Txt_s;

typedef Txt_s *    Txt;

/* what to to on database insert */
typedef enum {
   DBK_INSERT,   /* just insert */
   DBK_REPLACE,  /* replace entry with same key */
   DBK_MULTIPLE  /* allow several identical keys */
} Dbk_put_t;

#define DBK_F_EDITED       (1<<0)    /* dbk edited == unsorted   */
#define DBK_F_SORTR        (1<<1)    /* sortreverse, def = FALSE */
#define DBK_F_RDONLY       (1<<2)    /* readonly/locked, def = FALSE */
#define DBK_F_NOFORK       (1<<3)    /* Don't fork write, def = fork */

#define Dbk_getflag(dbk, f)    (((Dbk)dbk)->flags &   f)
#define Dbk_setflag(dbk, f)    (((Dbk)dbk)->flags |=  f)
#define Dbk_unsetflag(dbk, f)  (((Dbk)dbk)->flags &= !f)
#define Dbk_sort(dbk)  Dbk_sortkey((Dbk)dbk, ((Dbk)dbk)->keyo, ((Dbk)dbk)->keyl)
#define Dbk_cnt(dbk) ((Dbk)dbk)->recc
#define Dbk_get(dbk, key) Dbk_getkey(dbk, dbk->keyo, dbk->keyl, key)

typedef struct _Dbk_s {
   int        datas;     /* size of database*/
   char      *datap;     /* ptr to data */
      
   int        recl;      /* length of one record */
   int        recc;      /* no. of records */
   
   int        keyo;      /* keyoffset within record */
   int        keyl;      /* keylength */
   
   int        flags;     /* flags */
   long       mtime;     /* file last modified  */
   char      *filen;     /* Filename */
   
   int        chpid;     /* processid of save */
   
   void     (*proc)();   /* Error-proc */
} Dbk_s;

typedef Dbk_s* Dbk;

typedef enum {
   TLK_GET_MNI,
   TLK_GET_LABEL,
   TLK_GET_LABEL_P,
   TLK_GET_VENDOR
} Tlk_get_t;
   
typedef enum {
   TLK_SHOW,
   TLK_UNSHOW,
   
   TLK_ACTIVATE,
   TLK_DEACTIVATE,
   
   TLK_OPEN,
   TLK_CLOSE,
   
   TLK_INVERT,
   TLK_NORMALIZE,
   
   TLK_SET_DEFAULT_SCALE,
   TLK_SET_DEFAULT_FONT,
   TLK_SET_DEFAULT_COLOR,
   
   TLK_SET_XY,
   TLK_SET_LABEL,
   
   TLK_SET_CURSOR,
   
   TLK_SET_BORDER,
   TLK_UNSET_BORDER,
   
   TLK_APPEND_MNI,
   TLK_SET_MNU_ROWS,
         
   TLK_SET_MIDDLE_PROC,
   
   TLK_SET_TXT_MISC_PROC,
   TLK_SET_TXT_EDIT_PROC,
   
   TLK_SET_TXT,
   TLK_SET_ALL_TXT,
   TLK_SET_TXT_MODE,
   TLK_SET_TXT_STATUS,
   TLK_SET_TXT_CLEAR,
   
   TLK_SET_TXT_EDIT_FLAG,
   
   TLK_BELL,
   TLK_REDISP_ALL,
   
   TLK_SET_BUT_RPT
   
} Tlk_set_t;

#define DSP_DEFAULT_SCALE  -1
#define TLK_TIMEOUT         400

#define TLK_Adiaeresis      0xc4
#define TLK_Odiaeresis      0xd6
#define TLK_Udiaeresis      0xdc

#define TLK_adiaeresis      0xe4
#define TLK_odiaeresis      0xf6
#define TLK_udiaeresis      0xfc

#define TLK_ssharp          0xdf

#ifndef   _Tlk_

extern int   Dsp_rw[];      /* Display Raster Width */
extern int   Dsp_rh[];      /* Display Raster Height*/

extern GC   Dspgc;
extern GC   Dspgc_clr;
extern GC   Dspgc_inv;
extern GC   Dspgc_dbl;
extern GC   Dspgc_25;
extern GC   Dspgc_50;
extern GC   Dspgc_75;

extern Display  *Dsp;
extern int       Dsp_screen;
extern int       Dsp_slow;
extern int       Dsp_color;
extern int       Dsp_grey25;
extern int       Dsp_grey50;
extern int       Dsp_grey75;

extern int   Win_lastx;
extern int   Win_lasty;

extern char   *Win_io;
extern char   *Win_uml;
extern char   *Win_buf;
extern void Win_exit();

extern Win Win_create(
   Win,             /* Parent */
   Win_t,           /* Type */
   int,             /* Row */
   int,             /* Col */
   int,             /* Rows */
   int,             /* Cols */
   int,             /* Scale */
   char *,          /* Icon_bits */
   char *,          /* Label */
   void (*proc)()   /* Quit-Proc */
);

extern char * Win_malloc(
   int             /* Size */
);

extern char * Win_realloc(
   void *,        /* old-ptr */
   int             /* Size */
);

extern void Win_loop(
   Win             /* BaseWin */
);

extern void Win_set_middle_array(
   Win,         /* Window */
   int,         /* row */
   int,         /* col */
   int          /* len */
);

extern But But_create(
   Win,            /* Parent */
   int,            /* Row */
   int,            /* Col */
   char *,         /* Label */
   Mnu,            /* Menu */
   void (*proc)()  /* Action-Proc */
);

extern Lst Lst_create(
   Win,            /* Parent */
   int,            /* Row */
   int,            /* Col */
   int,            /* Rows */
   int,            /* Cols */
   void (*proc)(), /* Action-Proc */
   char *          /* Label */
);

extern Mnu Mnu_create(
   char **,        /* Labels */
   int,            /* Rows */
   int,            /* Scale */
   char*           /* Headline */
);

extern Txt Txt_create(
   Win,            /* Parent */
   int,            /* Row */
   int,            /* Col */
   int,            /* Length */
   Txt_t,          /* Type */
   Mnu,            /* Menu */
   void (*proc)(), /* Action-Proc */
   char *          /* Init-data */
);

extern Dbk Dbk_create(
   int,           /* RecLen    */
   int,           /* KeyOffset */
   int,           /* KeyLen    */
   void (*proc)() /* Error-Proc */
);

extern Dbk Dbk_update(
   Dbk,           /* Dbk    */
   char *           /* filen */
);

extern int Dbk_put(
   Dbk,           /* database to work on */
   char*,         /* record to put to database */
   Dbk_put_t      /* mode of operatrion */
);

extern void      Mnu_map(
   Mnu,            /* Menu */
   void (*proc)()  /* Action-Proc */   
);

extern void      Mnu_append_item(
   Mnu,            /* menu to appendt item to */
   char*,          /* label of this menu item */
   Mnu             /* sub-menu if any; NULL = none */
);

extern void      Dsp_open(
   int *,           /*  *argc */
   char **,         /* **argv */
   char *           /* Prog-Name */
);

extern Win_alarm_but_t   Win_alarm(
   char *,            /* Button 1 */
   char *,            /* Button 2 */
   char *,            /* Button 3 */
   char *,            /* Headline */
   char *,            /* Line 1 */
   char *,            /* Line 2 */
   char *,            /* Line 3 */
   char *,            /* Line 4 */
   char *             /* Line 5 */
);

extern void      Win_dl(
   Win,            /* Window */
   GC,             /* GraphicContext */
   int, int,       /* row1, col1 */
   int, int,       /* row2, col2 */
   int, int        /* xoff, yoff */
);

extern void      Win_db(
   Win,            /* Window */
   GC,             /* GraphicContext */
   int, int,       /* row, col */
   int, int,       /* rows, cols */
   int, int        /* xoff, yoff */
);

extern void      Win_fb(
   Win,            /* Window */
   GC,             /* GraphicContext */
   int, int,       /* row, col */
   int, int,       /* rows, cols */
   int, int        /* xoff, yoff */
);

extern void      Win_bt(
   Win,            /* Window */
   GC,             /* GraphicContext */
   int, int,       /* row, col */
   int, int,       /* xoff, yoff */
   char *,         /* string */
   int             /* length */
);

extern void      Win_ft(
   Win,            /* Window */
   GC,             /* GraphicContext */
   int, int,       /* row, col */
   int, int,       /* xoff, yoff */
   char *,         /* string */
   int             /* length */
);

extern void      Win_dt(
   Win,            /* Window */
   GC,             /* GraphicContext */
   int, int,       /* row, col */
   int, int,       /* xoff, yoff */
   char *,         /* string */
   int             /* length */
);

extern void      Win_bs(
   Win,            /* Window */
   GC,             /* GraphicContext */
   int, int,       /* row, col */
   int, int,       /* xoff, yoff */
   char *          /* string */
);

extern void      Win_fs(
   Win,            /* Window */
   GC,             /* GraphicContext */
   int, int,       /* row, col */
   int, int,       /* xoff, yoff */
   char *          /* string */
);

extern void      Win_ds(
   Win,            /* Window */
   GC,             /* GraphicContext */
   int, int,       /* row, col */
   int, int,       /* xoff, yoff */
   char *          /* string */
);

extern void      Win_quit();

extern void      Win_set_itimer(
   int,            /* Initial Value */
   int,            /* Reload Value */
   void (*proc)()  /* Action-Proc */   
);

extern void    Tlk_set();
extern caddr_t Tlk_get();

#endif /*_Tlk_*/


#ifdef _Tlk_
#include "dsp.h"
#include "win.h"
#include "but.h"
#include "txt.h"
#include "mnu.h"
#include "lst.h"
#include "dbk.h"
#include "scb.h"
void    Tlk_set();
caddr_t Tlk_get();
#endif

#endif   /* _Tlk_h */
