/* 
 * x-tlk Toolkit - Copyright (c) Holger Pfaff
 * http://pfaff.dnsalias.net
 *
 * This software maybe used for any purpose provided the
 * above copyright notice is retained. It is supplied as is.
 * No warranty expressed or implied - Use at your own risk.
 */

#ifndef _Txt_c
#define _Txt_c
#define _Tlk_

#include <tlk.h>

void Txt_createcursor() {
  int w = TXT_CURSOR_WIDTH, h = TXT_CURSOR_HEIGHT;
  if(Txt_cursor) return;
  Txt_cursor = XCreatePixmap(Dsp, Dsp_root, w, h, Dsp_depth);
  XFillRectangle(Dsp, Txt_cursor, Dspgc, 0, 0, w, h);
}

void Txt_init() {
   Txts = Txt_next = (Txt) Win_malloc( sizeof(Txt_s) );
   Txts->next = NULL;  
   Txt_prev = NULL;  
}

Txt Txt_create(Win win, int row, int col, int len, int type, Mnu mnu, void (*proc)(), char *data) {
   Txt      ptr;

   /* Win valid ? */
   if(win->text == 0) return 0;
   
   /* create linked list for txt fields */
   if(Txt_prev) Txt_prev->next = Txt_next;
   
   ptr         = Txt_next;
   ptr->prev   = Txt_prev;
   ptr->next   = NULL;
   Txt_prev    = ptr;
   Txt_next    = (Txt) Win_malloc( sizeof(Txt_s) );
   ptr->wtype  = WIN_TXT_WIN;
   ptr->win    = win;
   ptr->row    = row;
   ptr->col    = col;
   ptr->len    = len;
   ptr->flag   = FALSE;
   ptr->mnu    = mnu;
   ptr->proc   = proc;
   ptr->data   = win->data + ( win->cols * row + col );
   ptr->type   = (type < 2) ? TXT_TYPE_NUM  : TXT_TYPE_ALPHA;
   ptr->status = (type % 2) ? TXT_STATUS_RW : TXT_STATUS_RO;

   if(Txt_clear) {
      if(Txt_color) {
         Win_3d(win->pixmap, Dspgc_75, Dspgc_25, Dspgc_50, 
                col * Dsp_rw[win->scale] - 1, 
                row * Dsp_rh[win->scale] - 1, 
                len * Dsp_rw[win->scale] + 1, 
                  1 * Dsp_rh[win->scale] + 1 );
      } else {
         Win_fb(win, Dspgc_clr, row, col, 1, len + 1, -Dsp_rw[win->scale]/2, 0);
         Win_db(win, Dspgc,     row, col, 1, len + 1, -Dsp_rw[win->scale]/2, 0);
      }
   }      

   /* initialize field */
   while(len) *(win->text + (win->cols * row + col + --len)) = (void *) ptr;

   /* display desired string */
   Txt_set_txt(ptr, data);
            
   return ptr ;
}   

void Txt_button_press(Win win, XButtonEvent *xb) {
   Txt      txt;
   int      row, col;
   
   if(Txt_noedits) return; /* global no-edit ? */

   row = xb->y / Dsp_rh[win->scale]; 
   col = xb->x / Dsp_rw[win->scale];

   Txt_unset_cursor();
   
   /* is there a txt field? No -> try one further right */
   txt = (Txt)(win->text == 0 ? 0 : *(win->text + (win->cols * row + col)));
   if(txt == 0) txt = (Txt)(win->text == 0 ? 0 : *(win->text + (win->cols * row + col + 1)));

   /* left an txt field ? Call proc */
   if(Txt_active != txt  && Txt_active && Txt_active->proc) {
      Win_disable_timer();
      (void) (*Txt_active->proc)( Txt_active, 0, -1);
      Win_enable_timer();
   }

   Txt_active = txt;
   if(Txt_active == 0)  return; /* no txt field here ... */

   /* disable old selection if any */
   if(Txt_sel_sel) {            
      Win_ft(Txt_sel_txt->win, Dspgc, 
             Txt_sel_txt->row, Txt_sel_txt->col, 0, 0, 
             Txt_sel_txt->data, 
             Txt_sel_txt->len);
      Txt_sel_sel = FALSE;
   } 

   switch(xb->button) {
      case Button1: /* left mouse button */
            /* double click? Invert whole string */
            if(Txt_sel_lst  == Txt_active &&
               xb->time - Txt_sel_tme < TLK_TIMEOUT &&
               Txt_active->type != TXT_TYPE_PASSW) {
               /* find end of string */
               Txt_set_position(Txt_active->len + Txt_active->col - 1);
               /* no selections on empty fields */
               if(Txt_position + 1) {
                  Txt_sel_pos  = 0;
                  Txt_sel_len  = Txt_position + 1;
                  Txt_sel_txt  = Txt_active;
                  Txt_sel_sel  = TRUE;
               }
            /* partly selection using <shift> */
            } else if(Txt_sel_lst == Txt_active &&
                    Txt_active->type != TXT_TYPE_PASSW &&
                    Txt_shifted) {
               Txt_sel_pos  = Txt_position + 1;
               Txt_sel_len  = (col - Txt_active->col) - Txt_sel_pos + 1;
               Txt_sel_txt  = Txt_active;
               /* selection left of cursor ? */
               if(Txt_sel_len < 0) {
                  Txt_sel_pos += Txt_sel_len;
                  Txt_sel_len  = abs(Txt_sel_len);
               }
               /* check if something was selected */
               if(Txt_sel_len) Txt_sel_sel = TRUE;
            } else {
               /* no double click: just postion cursor */
               Txt_set_position(col);
               Txt_sel_tme = xb->time;
               Txt_sel_lst = Txt_active; 
            }
            /* invert selected text */
            if(Txt_sel_sel) {
               XSetSelectionOwner(Dsp, XA_PRIMARY, Txt_active->win->window, CurrentTime);
               Win_ft(Txt_sel_txt->win, Dspgc_inv, 
                      Txt_sel_txt->row, Txt_sel_txt->col + Txt_sel_pos, 0, 0, 
                      Txt_sel_txt->data + Txt_sel_pos, 
                      Txt_sel_len);
            }   
            Txt_set_cursor();
            break;
      case Button2: /* middle mouse button */
            /* get stuff from clipboard */
            if(*(Win_c->middle + (Win_c->cols * row + col))) {
               if(Win_middle_proc) {
                  Win_disable_timer();
                 (*Win_middle_proc) (row, col);
                  Win_enable_timer();
                  break;
               }
            }
            XConvertSelection(Dsp, XA_PRIMARY, XA_STRING, XA_STRING, 
                              Txt_active->win->window, CurrentTime);                              
            Txt_set_position(col);
            break;
      case Button3: /* right mouse button */
            /* show menu if any */
            if(txt->mnu) {
               Txt_sel_lst = Txt_active;
               Win_ft(txt->win,  Dspgc_inv, 
                      txt->row,  txt->col, 0, 0, 
                      txt->data, 
                      txt->len );
               Mnu_map(txt->mnu, Txt_mnu_proc);
               Txt_set_position(Txt_active->len + Txt_active->col - 1);
            }
            break;
   }
}

void Txt_key_press(KeySym keysym, char *chr, int len, int type) {
   int  del = 1;
   Txt  txt;
   if(Txt_misc_proc) {
       Win_disable_timer();
      (void)(*Txt_misc_proc)(Win_c, keysym, type);
       Win_enable_timer();
   }
   if(Mnu_act >= 0) {
      But_active = NULL;
      if(keysym == XK_Escape) Mnu_unmap(0);
      return;
   }
   if(Txt_active == 0 || type == KeyRelease) return;
   Txt_unset_cursor();
   txt = Txt_active;
   switch(keysym) {
      case XK_L8:        /* paste selection */
      case XK_Insert:
               XConvertSelection(Dsp, XA_PRIMARY, XA_STRING, XA_STRING, Txt_active->win->window, CurrentTime);                              
               break;
      case XK_L10:       /* delete selection, if local */
      case XK_Delete:
               if(Txt_sel_sel && txt->status == TXT_STATUS_RW) {
                  Txt_position = Txt_sel_pos + Txt_sel_len - 1;
                  del = Txt_sel_len;                                             
                  Txt_sel_len = Txt_sel_sel = 0;
               } else if(keysym != XK_Delete) {
                  break;
               }
      case XK_BackSpace:
               if(Txt_position == -1 || Txt_active->status == TXT_STATUS_RO) break;
               if(Txt_mode == TXT_MODE_INS) {
                  memcpy(Win_io, Txt_active->data + Txt_position + 1, Txt_active->len - Txt_position - 1);
                  memcpy(Txt_active->data + Txt_position - del + 1, Win_io, Txt_active->len - Txt_position - 1);
                  memset(Txt_active->data + Txt_active->len - del, ' ', del);
               } else {
                  memset(Txt_active->data + Txt_position - del + 1, ' ', del);
               }
               Txt_position -= del;      
               goto Txt_key_press_disp;
      case XK_Return:
      case XK_Linefeed:
      case XK_KP_Enter:
      case XK_Tab:
               if(Txt_sel_sel) {
                  Win_ft(Txt_sel_txt->win, Dspgc, 
                         Txt_sel_txt->row, Txt_sel_txt->col, 0, 0, 
                         Txt_sel_txt->data, 
                         Txt_sel_txt->len);
                  Txt_sel_sel = FALSE;
               }
               if(Txt_active->proc && Txt_active->status != TXT_STATUS_RO) (void) (*Txt_active->proc)(Txt_active, 0, -1);
               Txt_active = Txt_shifted ? Txt_active->prev : Txt_active->next;
               if(Txt_active == 0) Txt_active = Txts; 
               Txt_set_position(Txt_active->len + Txt_active->col - 1);
               if(Txt_edited == 0 && Txt_edited_proc && Txt_active->status != TXT_STATUS_RO) {
                  Txt_edited = TRUE;
                  (void) (*Txt_edited_proc)(Txt_active);
               }
               Txt_sel_pos  = 0;
               Txt_sel_len  = Txt_position + 1;
               Txt_sel_txt  = Txt_active;
               Txt_sel_sel  = TRUE;
               XSetSelectionOwner(Dsp, XA_PRIMARY, Txt_active->win->window, CurrentTime);
               Win_ft(Txt_sel_txt->win, Dspgc_inv, 
                      Txt_sel_txt->row, Txt_sel_txt->col + Txt_sel_pos, 0, 0, 
                      Txt_sel_txt->data + Txt_sel_pos, 
                      Txt_sel_len);
               Txt_set_cursor();
               return;
      case XK_Right:
               if(Txt_position <  txt->len - 1) {
                  ++Txt_position;
               } else {
                  if(Txt_active->proc && Txt_active->status != TXT_STATUS_RO) (void) (*Txt_active->proc)(Txt_active, 0, -1);
                  Txt_active = Txt_active->next;
                  if(Txt_active == 0) Txt_active = Txts;
                  Txt_set_position(Txt_active->len + Txt_active->col);
               }
               break;
      case XK_Left:
               if(Txt_position >= 0) {
                 --Txt_position;
               } else if(Txt_active->prev) {
                  if(Txt_active->proc && Txt_active->status != TXT_STATUS_RO) (void) (*Txt_active->proc)(Txt_active, 0, -1);
                  Txt_active = Txt_active->prev;
                  Txt_set_position(Txt_active->len + Txt_active->col);
               }           
               break;
      case XK_Up:
      case XK_Down:
               if(Txt_active->proc && Txt_active->status != TXT_STATUS_RO ) (void) (*Txt_active->proc)(Txt_active, 0, -1);
               if(Txt_active->row <= 1 || Txt_active->row >= Txt_active->win->rows - 1) break;
               txt = (Txt) *(Txt_active->win->text + 
                            (Txt_active->win->cols * 
                            (Txt_active->row + (keysym == XK_Down ? 1 : -1)) + 
                             Txt_active->col));
               if(txt != Txt_active && txt) Txt_active = txt;
               Txt_set_position(Txt_active->len + Txt_active->col);
               break;
      default:
               if(Txt_active->status == TXT_STATUS_RO || len == 0) break;
               if(Txt_active->type == TXT_TYPE_ALPHA && Txt_is_alpha(chr, len) == 0) break;        
               if(Txt_active->type == TXT_TYPE_NUM && Txt_is_num(chr, len) == 0) break;  
               if(Txt_sel_sel) {
                  Txt_key_press(XK_Delete, chr, len, type);
                  Txt_unset_cursor();
               }
               if(Txt_position >= Txt_active->len - 1) break;
               len = len > Txt_active->len - Txt_position - 1 ? Txt_active->len - Txt_position - 1 : len;
               if(Txt_mode == TXT_MODE_INS) {
                  memcpy(Win_io, Txt_active->data + Txt_position + 1, Txt_active->len - Txt_position - 1);
                  memcpy(Txt_active->data + Txt_position + 1, chr, len);
                  memcpy(Txt_active->data + Txt_position + len + 1, Win_io, Txt_active->len - Txt_position - 1 - len);
               } else {
                  memcpy(Txt_active->data + Txt_position + 1, chr, len);
               }
               Txt_position += len;
Txt_key_press_disp: 
               del = Txt_position - len;
               del = del < 0 ? 0 : del;
               if(Txt_active->type == TXT_TYPE_PASSW) {
                  Win_ft(Txt_active->win, Dspgc, Txt_active->row, Txt_active->col, 0, 0, "        ", Txt_active->len);
                  Win_ft(Txt_active->win, Dspgc, Txt_active->row, Txt_active->col, 0, 0, "********", strlen((char *)Tlk_get(Txt_active, TLK_GET_LABEL)));
               } else {
                  Win_ft(Txt_active->win, Dspgc, Txt_active->row, Txt_active->col + del, 0, 0, Txt_active->data + del, Txt_active->len  - del);
               }
               if(Txt_edited == 0 && Txt_edited_proc) {
                  Txt_edited = TRUE;
                 (void) (*Txt_edited_proc)(Txt_active);
               }
               break;    
   }
   if(Txt_sel_sel) {
      Win_ft( Txt_sel_txt->win, Dspgc, Txt_sel_txt->row, Txt_sel_txt->col, 0, 0, Txt_sel_txt->data, Txt_sel_txt->len);
      Txt_sel_sel = FALSE;
   }
   Txt_set_cursor();
}

/*
 * check if char is alpha - maps \n\r\t to ' '
 * recognizes german umlauts
 */
int Txt_is_alpha(unsigned char *chr, int len) {
   while(len--) {
      if(chr[len] == '\n' || 
         chr[len] == '\r' ||
         chr[len] == '\t'   )
         
         chr[len] = ' ';
      
      if(chr[len] < ' ') return FALSE;
         
      if(chr[len] > '~'  &&
         !(chr[len] == TLK_Adiaeresis ||
           chr[len] == TLK_Odiaeresis ||
           chr[len] == TLK_Udiaeresis ||
           chr[len] == TLK_adiaeresis ||
           chr[len] == TLK_odiaeresis ||
           chr[len] == TLK_udiaeresis ||
           chr[len] == TLK_ssharp   ) )
         return FALSE;
   }
   return TRUE;   
}

/*
 * check if char is numerical - includes -,.
 */
int Txt_is_num(unsigned char *chr, int len) {
   while(len--) {
      chr[len] = chr[len] == ',' ? '.' : chr[len];
      if(isdigit(chr[len]) == 0 && chr[len] != '.' && chr[len] != ' ' && chr[len] != '-') return FALSE;
   } 
   return TRUE;
}

/*
 * set cursor to pos col - tries to find a reasonable
 * non-blank position
 */
void Txt_set_position(int col) {

   if(Txt_active == NULL) return;

   Txt_position = col - Txt_active->col;
   col = Txt_position;

   if(*(Txt_active->data + col) == ' ') {
      while(col < Txt_active->len) if(*(Txt_active->data + col) == ' ') ++col; else return;
      while(Txt_position >= 0) if(*(Txt_active->data + Txt_position) == ' ') --Txt_position; else return;         
   }
}

void Txt_set_cursor() {
   if(Txt_active == 0 || Txt_active->status == TXT_STATUS_RO) return;

   Txt_createcursor();

   XCopyArea(Dsp, Txt_cursor, Txt_active->win->window, Dspgc, 0, 0, TXT_CURSOR_WIDTH, TXT_CURSOR_HEIGHT, 
            (Txt_active->col + Txt_position + 1) * Dsp_rw[Txt_active->win->scale] - (TXT_CURSOR_WIDTH + 1) / 2,
            (Txt_active->row + 1               ) * Dsp_rh[Txt_active->win->scale] -  TXT_CURSOR_HEIGHT/ 2 -1);
}

void Txt_unset_cursor() {
   int x, y;

   if(Txt_active == 0) return;

   Txt_createcursor();

   x = (Txt_active->col + Txt_position + 1) * Dsp_rw[Txt_active->win->scale] - (TXT_CURSOR_WIDTH  + 1)/ 2;
   y = (Txt_active->row + 1)                * Dsp_rh[Txt_active->win->scale]  - TXT_CURSOR_HEIGHT / 2 - 1;
        
   XCopyArea(Dsp, Txt_active->win->pixmap, Txt_active->win->window, Dspgc, x, y, TXT_CURSOR_WIDTH, TXT_CURSOR_HEIGHT, x, y);
}

void Txt_mnu_proc(Mnu mnu, char *label, int item) {
   if(Txt_active == 0) return;
   if(item == -1) return;
      
   Txt_set_txt(Txt_active, label);
   Txt_unset_cursor();
   Txt_set_position(Txt_active->len + Txt_active->col);

   if(Txt_edited == 0 && Txt_edited_proc) {
      Txt_edited = TRUE;
      (void) (*Txt_edited_proc)(Txt_active);
   }

   if(Txt_active->proc) (void) (*Txt_active->proc)(Txt_active, label, item);
}

/*
 * Show string or clear field (str=NULL)
 */
void Txt_set_txt(Txt txt, char *str) {
   if(str == 0 || strcmp(str, "-") == 0) {
      if(strlen((char *)Tlk_get((Win)txt, TLK_GET_LABEL, 0)) || txt->type == TXT_TYPE_PASSW) {
         memset(txt->data, ' ', txt->len);
      } else {
         return;
      }
   } else {
      if(memcmp(txt->data, str, txt->len)) {
         memset(txt->data, ' ', txt->len),
         memcpy(txt->data, str, strlen(str) > txt->len ? txt->len : strlen(str) );   
      } else {
         return;
      }
   }   
   if(txt->type == TXT_TYPE_PASSW && str) {
      Win_ft(txt->win, Dspgc, txt->row, txt->col, 0, 0, "********", strlen((char *) Tlk_get(txt, TLK_GET_LABEL)));
   } else {
      Win_ft(txt->win, Dspgc, txt->row, txt->col, 0, 0, txt->data, txt->len);
   }
}

#endif /* _Txt_c */
