/* 
 * x-tlk Toolkit - Copyright (c) Holger Pfaff
 * http://pfaff.dnsalias.net
 *
 * This software maybe used for any purpose provided the
 * above copyright notice is retained. It is supplied as is.
 * No warranty expressed or implied - Use at your own risk.
 */

#ifndef _Win_c
#define _Win_c
#define _Tlk_

#include <tlk.h>

void Win_init() {
   int  i;

   for(i = 0;i < WIN_FACTOR; ++i) Wins[i] = Win_last[i] = NULL;

   Win_io    = (char *) Win_malloc(3*WIN_IO_BUFSIZE+WIN_LABEL_LEN);
   Win_buf   = Win_io + WIN_IO_BUFSIZE;
   Win_uml   = Win_buf+ WIN_IO_BUFSIZE;
   Win_label = Win_uml+ WIN_IO_BUFSIZE;
   
   Win_atoms[0] = XInternAtom(Dsp, "WM_DELETE_WINDOW", 0);
   Win_atoms[2] = XInternAtom(Dsp, "WM_PROTOCOLS", 0);
   Win_atoms[1] = 0;
}

Win Win_create(Win par, Win_t type, int row, int col, int rows, int cols, int sc, unsigned char *icon_bits, char *label, void (*proc)()) {
   XSizeHints           size_hints;
   XWMHints             wm_hints;
   XSetWindowAttributes win_attr;

   Pixmap          icon;
   long            mask;

   int             border = 0;
   Win             ptr = (Win) Win_malloc(sizeof(Win_s));
   
   ptr->scale = type == WIN_MNI_WIN ? par->scale : Dsp_check_scale(sc);

   ptr->x = col * Dsp_rw[ptr->scale];
   ptr->y = row * Dsp_rh[ptr->scale];

   ptr->w = cols * Dsp_rw[ptr->scale];
   ptr->h = rows * Dsp_rh[ptr->scale];

   ptr->cols = cols++;
   ptr->rows = rows;
   ptr->type = type;

   ptr->label = label;
   ptr->proc  = proc;

   ptr->next   = 0;
   ptr->middle = 0;
   ptr->text   = 0;
   ptr->pixmap = 0;
   ptr->data   = 0;
   ptr->flags  = 0;

   ptr->parent = par;
   ptr->cnt    = 0;

   ptr->window = XCreateSimpleWindow(Dsp,
                 par == 0 || type == WIN_POP_WIN ? Dsp_root : par->window,
                 ptr->x, ptr->y, ptr->w, ptr->h, 0,
                 Dsp_black, Dsp_white);

   if(Win_last[ptr->window % WIN_FACTOR] == 0) {
      Wins[ptr->window % WIN_FACTOR] = ptr;
   } else {
      Win_last[ptr->window % WIN_FACTOR]->next = ptr;
   }

   Win_last[ptr->window % WIN_FACTOR] = ptr;

   win_attr.override_redirect = TRUE;
   win_attr.save_under = FALSE;

   switch(type) {
      case WIN_BAS_WIN:
      case WIN_POP_WIN:
      case WIN_SUB_WIN:
         ptr->middle = (int *)   Win_malloc(rows * cols * sizeof(int));
         ptr->text   = (void **) Win_malloc(rows * cols * sizeof(int));
         ptr->data   = (char *)  Win_malloc(rows * cols + rows);

         memset((char *) ptr->data, ' ', rows * cols + rows);
         memset((char *) ptr->text, 0, rows * cols + sizeof(int));
         memset((char *) ptr->middle, 0, rows * cols + sizeof(int));
   
         mask = KeyPressMask | ButtonPressMask | ButtonReleaseMask | ExposureMask | KeyReleaseMask; 
         ptr->pixmap = XCreatePixmap(Dsp, ptr->window, ptr->w, ptr->h, Dsp_depth);
         
         Win_fb(ptr, (Dsp_color && type != WIN_BAS_WIN) || Txt_color ? Dspgc_50 : Dspgc_clr, 0, 0, rows, --cols, 0, 0);
   
         if(Txt_color) XSetWindowBackground(Dsp, ptr->window, Dsp_grey50);
   
         if(type  == WIN_SUB_WIN) {
            win_attr.save_under = TRUE;
            border = 4;
            break;
         } else {
            border = 2;
         }
         if(type  == WIN_POP_WIN) XSetTransientForHint(Dsp, ptr->window, par->window);
         size_hints.max_width  = ptr->w;
         size_hints.max_height = ptr->h;
         size_hints.flags = PMaxSize;
   
         wm_hints.initial_state = Win_iconic ? IconicState : NormalState;
         wm_hints.input  = TRUE;
         wm_hints.flags  = StateHint | InputHint;
   
         if(Win_ix >= 0 && Win_iy >= 0) {
            wm_hints.icon_x = Win_ix;
            wm_hints.icon_y = Win_iy;
            wm_hints.flags |= IconPositionHint;
         }
         if(icon_bits) {
            icon = XCreatePixmapFromBitmapData(Dsp, (Drawable)ptr->window, (char *) icon_bits, (unsigned int) 64, (unsigned int) 45, 
                                               (unsigned long) Dsp_black, (unsigned long) Dsp_white, (unsigned int) 1);
            wm_hints.flags      |= IconPixmapHint;
            wm_hints.icon_pixmap = icon;
         }
         memset(Win_label,' ', WIN_LABEL_LEN);
         Win_label[WIN_LABEL_LEN] = '\0';
         
         if(label) strncpy(Win_label, label, WIN_LABEL_LEN);
   
         XSetStandardProperties(Dsp, ptr->window, label, Win_label, (Pixmap) icon, (char **) NULL, 0, &size_hints);
         XSetWMHints(Dsp, ptr->window, &wm_hints);
         XChangeProperty(Dsp, ptr->window, Win_atoms[2], XA_ATOM, 32, PropModeReplace, (unsigned char *) Win_atoms, 2);
         XDefineCursor(Dsp, ptr->window, Dspcur[(int)DSPCUR_NRM]);
         win_attr.override_redirect = FALSE;
         break;
   
      case WIN_MNI_WIN:
         XSetWindowBorder(Dsp, ptr->window, Dsp_white);
   
      case WIN_BUT_WIN:
         border = (type == WIN_BUT_WIN && Dsp_color) ? 0 : 1;
         XResizeWindow(Dsp, ptr->window, ptr->w - border, ptr->h - border);
         if(type == WIN_BUT_WIN) XDefineCursor(Dsp, ptr->window, Dspcur[(int)DSPCUR_HND]);
         if(Dsp_color) {
            XSetWindowBackground(Dsp, ptr->window, Dsp_grey50);
            if(type == WIN_BUT_WIN) {
               XMoveResizeWindow(Dsp, ptr->window, ptr->x, ptr->y - 3, ptr->w, ptr->h += 6);
            } else {
               XSetWindowBorder(Dsp, ptr->window, Dsp_grey50);
            }
         }
         mask = EnterWindowMask | ButtonPressMask   | LeaveWindowMask | ButtonReleaseMask | ExposureMask | OwnerGrabButtonMask; 
         XMapRaised(Dsp, ptr->window);
         break;
   
      case WIN_MNB_WIN:
         mask = ExposureMask;
         win_attr.save_under = TRUE;
         border = 1;
         if(Dsp_color) XSetWindowBackground(Dsp, ptr->window, Dsp_grey50);
         break;
         
      case WIN_LST_WIN:
         mask = ButtonPressMask | ExposureMask | KeyPressMask | ButtonReleaseMask;
         Win_fb(ptr, Dspgc_clr, 0, 0, ptr->rows, ptr->cols, 0, 0);
         if(Dsp_color) XSetWindowBackground(Dsp, ptr->window, Dsp_grey50);
         border = 0;
         XMapRaised(Dsp, ptr->window);
         break;
   
      case WIN_ALB_WIN:
         mask = ExposureMask;
         win_attr.save_under = TRUE;
         if(Dsp_color) XSetWindowBackground(Dsp, ptr->window, Dsp_grey50);
         border = Dsp_color ? 1 : 3;
         ptr->pixmap = XCreatePixmap(Dsp, ptr->window, ptr->w, ptr->h, Dsp_depth);
         win_attr.save_under = FALSE;
      default:
         break;
   }
   XChangeWindowAttributes(Dsp, ptr->window, CWOverrideRedirect | CWSaveUnder, &win_attr);
   XSetWindowBorderWidth(Dsp, ptr->window, border);
   XSelectInput(Dsp, ptr->window, mask);
   return ptr;
}
   
void Win_loop(Win win) {
   XEvent          event;
   if(Win_alarm_win == 0) Win_alarm_create_win();

   if(win) {
      XMapRaised(Dsp, win->window);
      XSetSelectionOwner(Dsp, XA_PRIMARY, win->window, CurrentTime);
      if(Win_x >= 0 && Win_y >= 0) XMoveWindow(Dsp, win->window, Win_x, Win_y);
      if(Win_w && Win_h) XResizeWindow(Dsp, win->window, Win_w, Win_h);
   }
   for(;;) {
      XNextEvent(Dsp, &event);
      if(Win_events(&event) == FALSE) break;
   }
}

void Win_event() {
   XEvent event;
   while(XCheckMaskEvent(Dsp, ExposureMask, &event)) {
      Win_events(&event);
   }
}

int Win_events(XEvent *event) {
   KeySym          keysym;
   XComposeStatus  compose;
   int             chr_cnt;
   char            chr[3];
   unsigned char  *ptr;
   unsigned long   cnt, rest;
   Atom            atom;

   Win_lock = TRUE;
   Win_c = Wins[event->xexpose.window % WIN_FACTOR];
   while (Win_c) {
      if (Win_c->window == event->xexpose.window) goto Win_loop_cont;
      Win_c = Win_c->next;
   }
   return TRUE;
   
Win_loop_cont:
   switch (event->type) {
      case GraphicsExpose:
      case Expose:
         switch (Win_c->type) {
         case WIN_BUT_WIN:
         case WIN_MNI_WIN:
         case WIN_ABU_WIN:
         case WIN_RBU_WIN:
            But_draw(Win_c);
            break;
         case WIN_LST_WIN:
            Lst_dsp(Win_c);
            break;
         case WIN_SCB_WIN:
            Scb_dsp(Win_c);
            break;
         case WIN_MNB_WIN:
            if(Dsp_color) Win_3d(Win_c->window, Dspgc_25, Dspgc_75, Dspgc_50, 2, 2, Win_c->w - 5, Win_c->h - 5);
            if(Win_c->label) Win_bs(Win_c, Dspgc, 0, Win_c->cols / 2 - strlen(Win_c->label) / 2, Dsp_color ? 2 : 0, Dsp_color ? 2 : 0, Win_c->label);
            break;
         default:
            if(Win_c->pixmap) {
               XCopyArea(Dsp, Win_c->pixmap, Win_c->window, DefaultGC(Dsp, Dsp_screen), event->xexpose.x, event->xexpose.y,
                         event->xexpose.width, event->xexpose.height, event->xexpose.x, event->xexpose.y);
               Txt_set_cursor();
            }
            break;
         }
         break;
      case EnterNotify:
      case LeaveNotify:
         Win_enter_leave(&event->xcrossing);
         break;
      case ButtonPress:
         Win_button_press(&event->xbutton);
         break;
      case ButtonRelease:
         Win_button_release(&event->xbutton);
         if (Win_alarm_active && Win_alarm_but) return FALSE;
         break;
      case KeyPress:
      case KeyRelease:
         chr_cnt = XLookupString(&event->xkey, chr, 2, &keysym, &compose);
         if(Dsp_kbdtype) {
            XKeyboardControl kbcontrol;
            XKeyboardState kbstate;
            int led_mask  = Dsp_is_ow ? 0x00000001 : 0x00000004;
            kbcontrol.led = Dsp_is_ow ? 1 : 3;
            XGetKeyboardControl(Dsp, &kbstate);
            
            if(event->xkey.keycode == 105) {
               if(Dsp_is_ow == FALSE && event->type == KeyRelease) {
                  kbcontrol.led_mode = (kbstate.led_mask & led_mask) ? LedModeOff : LedModeOn;
                  XChangeKeyboardControl(Dsp, KBLed|KBLedMode, &kbcontrol);
               }
               break;
            }
            if(kbstate.led_mask & led_mask) {
               /* Numlock is on */
               switch(event->xkey.keycode) {
                  case 101: keysym = XK_0;      chr[0] = '0';chr_cnt = 1; break;
                  case  57: keysym = XK_period; chr[0] = '.';chr_cnt = 1; break;
                  case 119: keysym = XK_1;      chr[0] = '1';chr_cnt = 1; break;
                  case 120: keysym = XK_2;      chr[0] = '2';chr_cnt = 1; break;
                  case 121: keysym = XK_3;      chr[0] = '3';chr_cnt = 1; break;
                  case  98: keysym = XK_4;      chr[0] = '4';chr_cnt = 1; break;
                  case  99: keysym = XK_5;      chr[0] = '5';chr_cnt = 1; break;
                  case 100: keysym = XK_6;      chr[0] = '6';chr_cnt = 1; break;
                  case  75: keysym = XK_7;      chr[0] = '7';chr_cnt = 1; break;
                  case  76: keysym = XK_8;      chr[0] = '8';chr_cnt = 1; break;
                  case  77: keysym = XK_9;      chr[0] = '9';chr_cnt = 1; break;
               }
            } else {
               /* Numlock is off */
               switch(event->xkey.keycode) {
                  case 101: keysym = XK_Insert; chr[0] =  0 ;chr_cnt = 0; break;
                  case  57: keysym = XK_Delete; chr[0] =  0 ;chr_cnt = 0; break;
                  case 119: keysym = XK_End;    chr[0] =  0 ;chr_cnt = 0; break;
                  case 120: keysym = XK_Down;   chr[0] =  0 ;chr_cnt = 0; break;
                  case 121: keysym = XK_Next;   chr[0] =  0 ;chr_cnt = 0; break;
                  case  98: keysym = XK_Left;   chr[0] =  0 ;chr_cnt = 0; break;
                  case  99: keysym = XK_5;      chr[0] = '5';chr_cnt = 1; break;
                  case 100: keysym = XK_Right;  chr[0] =  0 ;chr_cnt = 0; break;
                  case  75: keysym = XK_Home;   chr[0] =  0 ;chr_cnt = 0; break;
                  case  76: keysym = XK_Up;     chr[0] =  0 ;chr_cnt = 0; break;
                  case  77: keysym = XK_Prior;  chr[0] =  0 ;chr_cnt = 0; break;
               }
            }
         }   
         if (keysym == XK_Shift_L || keysym == XK_Shift_R) {
            Txt_shifted = (event->type == KeyPress) ? TRUE : FALSE;
            break;
         }
         switch(Win_c->type) {
            case WIN_LST_WIN:
               Lst_key_press(Win_c, keysym, chr, chr_cnt);
               break;
            default:
               Txt_key_press(keysym, chr, chr_cnt, event->type);
               break;
         }
         break;
      case ConfigureNotify:
         switch(Win_c->type) {
            case WIN_BAS_WIN:
            case WIN_POP_WIN:
               Win_c->x = event->xconfigure.x;
               Win_c->y = event->xconfigure.y;
               Win_c->w = event->xconfigure.width;
               Win_c->h = event->xconfigure.height;
            default:
               break;
         }
         break;
      case VisibilityNotify:
         XMapRaised(Dsp, Win_alarm_win->window);
         break;
      case ClientMessage:
         if(event->xclient.message_type == Win_atoms[2] && event->xclient.data.l[0] == Win_atoms[0]) {
            if(Win_c->proc) {
               Win_disable_timer();
               Win_lastx = DisplayWidth(Dsp, Dsp_screen) / 2;
               Win_lasty = DisplayHeight(Dsp, Dsp_screen) / 2;
              (void) (*Win_c->proc) (Win_c, Win_c->label);
               Win_enable_timer();
            } else if(Win_c->parent == 0) {
               Win_quit();
            } else {
               Tlk_set(Win_c, TLK_UNSHOW);
            }
         }
         break;
      case SelectionClear:
         /* Someone wants to become selection owner */
         if(Txt_sel_sel) Win_ft(Txt_sel_txt->win, Dspgc, Txt_sel_txt->row, Txt_sel_txt->col, 0, 0, Txt_sel_txt->data, Txt_sel_txt->len);
         Txt_sel_tme = (Time) 0;
         Txt_sel_sel = FALSE;
         Txt_sel_lst = NULL;
         Txt_sel_txt = NULL;
         break;
      case SelectionRequest:
         /* soneone wants our selection data */
         if(Txt_sel_txt) {
            XEvent          svt;
            svt.type = SelectionNotify;
            svt.xselection.send_event = TRUE;
            svt.xselection.display = event->xselectionrequest.display;
            svt.xselection.requestor = event->xselectionrequest.requestor;
            svt.xselection.selection = event->xselectionrequest.selection;
            svt.xselection.target = event->xselectionrequest.target;
            svt.xselection.property = event->xselectionrequest.property;
            svt.xselection.time = event->xselectionrequest.time;
            XChangeProperty(Dsp, svt.xselection.requestor, svt.xselection.property, svt.xselection.target, 8,
                            PropModeReplace, (unsigned char *) Txt_sel_txt->data + Txt_sel_pos, Txt_sel_len);
            XSendEvent(svt.xselection.display, svt.xselection.requestor, FALSE, 0L, &svt);
         }
         break;
      case SelectionNotify:
         /* requested selection data arrives */
         cnt = chr_cnt = 0;
         if(Txt_active && event->xselection.target) {
            XGetWindowProperty(event->xselection.display, event->xselection.requestor, event->xselection.target,
                               0L, 100L, FALSE, XA_STRING, &atom, &chr_cnt, &cnt, &rest, &ptr);
            if (cnt && ptr) Txt_key_press(0, ptr, cnt);
         }
         break;
      default:
         fprintf(stderr, "%s: Unexpected event %d\n", Dsp_prog, event->type);
      case NoExpose:
      case MapNotify:
      case UnmapNotify:
      case ReparentNotify:
         break; /* ignore these events for now */
   }
   XFlush(Dsp);
   Win_lock = FALSE;
   return TRUE;
}

void Win_enter_leave(XCrossingEvent *xcrossing) {
   if(Mnu_act == -1) But_active = NULL;
   switch (Win_c->type) {
      case WIN_BUT_WIN:
         if(Mnu_act >= 0) break;
      case WIN_MNI_WIN:
         if(Win_alarm_active) break;
      case WIN_ABU_WIN:
         if(But_xbutton_pressed == 0) break;
         if(xcrossing->type == EnterNotify) {
            But_select(Win_c);
            if(Mnu_act >= 0 && (Mnu) Win_c->middle != Mnus[Mnu_act]) {
               while (Win_c->parent != Mnus[Mnu_act]) {
                  XUnmapWindow(Dsp, Mnus[Mnu_act--]->window);
               }
               if(Win_c->middle) {
                  Win_lastx = xcrossing->x_root;
                  Win_lasty = xcrossing->y_root;
                  Mnu_map(Win_c->middle, Win_c->parent->proc);
               }
            }
         } else {
            But_draw(Win_c);
         }
      default:
         break;
   }
}

void Win_sleep(int msec) {
   struct timeval timeout;
   timeout.tv_sec = msec > 999 ? msec/1000 : 0;
   msec *= 1000;
   timeout.tv_usec = msec % 1000000;
   select(0, 0, 0, 0, &timeout);
}

void Win_button_press(XButtonEvent *xbutton) {
   int          x;
   
   But_xbutton_pressed = TRUE;
   Win_lastx = xbutton->x_root;
   Win_lasty = xbutton->y_root;
   
   x = Win_lastx;

   switch(Win_c->type) {
      case WIN_BAS_WIN:
      case WIN_SUB_WIN:
      case WIN_POP_WIN:
         if(Mnu_act >= 0 || Win_alarm_active) {
            Mnu_item_selected = Mnu_stay_up = FALSE;
            break;
         }
         Txt_button_press(Win_c, xbutton);
         break;
   
      case WIN_SCB_WIN:
         Scb_button_press(Win_c, xbutton);
         break;
         
      case WIN_LST_WIN:
         if(Mnu_act >= 0 || Win_alarm_active) {
            Mnu_item_selected = Mnu_stay_up = FALSE;
            break;
         }
         Lst_button_press(Win_c, xbutton);
         break;
              
      case WIN_RBU_WIN:
         if(Txt_active) Txt_button_press(Win_c, xbutton);
         if(Mnu_act >= 0 || Win_alarm_active) {
            Mnu_item_selected = Mnu_stay_up = FALSE;
            break;
         }
         But_select(Win_c);
         if(Win_c->proc) {
            unsigned int mask;
            Window       rW,cW;
            int          rx, ry, lx, ly;
            
           (void)(*Win_c->proc)(Win_c, Win_c->label, xbutton->time);
                                    
            XFlush(Dsp);
            Win_sleep(150);
            Win_disable_timer();
         
            while(XQueryPointer(Dsp,Win_c->window,&rW,&cW,&rx,&ry,&lx,&ly,&mask)) {   
               if((mask & Button1Mask) == 0)  /* Button released */ {
                  But_xbutton_pressed = FALSE;
                  break; 
               }   
               if(lx < 0 || ly < 0 || lx > Win_c->w || ly > Win_c->h) break; /* Window left ? */
               (void)(*Win_c->proc)(Win_c, Win_c->label, xbutton->time);
               Win_sleep(50);
            }
            Win_enable_timer();
         }
         But_draw(Win_c);
         break;
      
      case WIN_BUT_WIN:
         if(Txt_active) Txt_button_press(Win_c, xbutton);
         if(Mnu_act >= 0 || Win_alarm_active) {
            Mnu_item_selected = Mnu_stay_up = FALSE;
            break;
         }
         if(Win_c->data && xbutton->button == Button3) Mnu_map(Win_c->data, Win_c->proc);
         But_active = Win_c;
   
      case WIN_MNI_WIN:
         if(Win_c->type == WIN_MNI_WIN && Win_c->middle) {
            while(Win_c->parent != Mnus[Mnu_act]) {
               XUnmapWindow(Dsp, Mnus[Mnu_act--]->window);
            }
            Mnu_map(Win_c->middle, Win_c->parent->proc);
         }
   
      case WIN_ABU_WIN:
         But_xbutton_pressed = TRUE;
         But_select(Win_c);
      default:
         break;
   }
}

void Win_button_release(XButtonEvent *xbutton) {
   But but;   

   switch (Win_c->type) {
      case WIN_BUT_WIN:
         if(Mnu_act >= 0 || Win_alarm_active) break;
         if(Win_c->data && ((Mnu) Win_c->data)->data) xbutton->time = 0;
         But_busy(but = Win_c);
         But_active = NULL;
         if(Win_c->proc) {
            Win_disable_timer();
            (void) (*Win_c->proc) (xbutton->time ? Win_c : (But) Win_c->data,
                         xbutton->time ? Win_c->label : ((Mnu) ((Mnu) Win_c->data)->data)->label,
                         xbutton->time);
            Win_enable_timer();
         }
         But_draw(but);
         break;
   
      case WIN_ABU_WIN:
         if(But_xbutton_pressed) (void)(*Win_c->proc)(Win_c);
         break;
   
      case WIN_MNI_WIN:
         if(Mnu_act == -1) {
            Win_disable_timer();
            (void) (*Win_c->proc) (xbutton->time ? Win_c : (But) Win_c->data,
                         xbutton->time ? Win_c->label : ((Mnu) ((Mnu) Win_c->data)->data)->label,
                         xbutton->time);
            Win_enable_timer();
         }
         Mnu_item_selected = TRUE;
         if(Win_c->middle == 0) Mnu_stay_up = FALSE;
         But_draw(Win_c);
         break;
         
      case WIN_MNB_WIN:
         Mnu_stay_up = FALSE;
            
      case WIN_BAS_WIN:
      case WIN_POP_WIN:
      case WIN_SUB_WIN:
         if(Win_lastx - xbutton->x_root > 2 || Win_lasty - xbutton->y_root > 2) Mnu_stay_up = FALSE;
      default:
         break;
   }
   
   But_xbutton_pressed = FALSE;
   Mnu_unmap(xbutton);
}

void Win_alarm_create_win() {
   int i = 0;
   
   Win_stop_pix = XCreatePixmapFromBitmapData(Dsp, Dsp_root, (char*)Win_stop_bits, 25, 25,
                  Dsp_black, Dsp_color ? Dsp_grey50 : Dsp_white, Dsp_depth);
   Win_alarm_win = Win_create(NULL, WIN_ALB_WIN, 0, 0, WIN_ALARM_HEIGTH, WIN_ALARM_WIDTH, 
                              DSP_DEFAULT_SCALE, 0, 0, 0);             
   for(; i < 3; ++i) {
      Win_alarm_buts[i] = But_create(Win_alarm_win, WIN_ALARM_BUT_ROW,
                                     WIN_ALARM_BUT_COL(i), WIN_ALARM_EMPTY,
                                     0, Win_alarm_proc);
      Win_alarm_buts[i]->type = WIN_ABU_WIN;
   }
   Win_lastx = DisplayWidth(Dsp, Dsp_screen)/2;
   Win_lasty = DisplayHeight(Dsp, Dsp_screen)/2;
}

Win_alarm_but_t Win_alarm(char *but1, char *but2, char *but3, char *hl, char *l1, char *l2, char *l3, char *l4, char *l5) {
   char *lines[5];
   int   i, x;

   if(Win_alarm_win == 0) Win_alarm_create_win();

   Win_alarm_buts[0]->label = but1;
   Win_alarm_buts[1]->label = but2;
   Win_alarm_buts[2]->label = but3;

   lines[0] = l1;
   lines[1] = l2;
   lines[2] = l3;
   lines[3] = l4;
   lines[4] = l5;

   if((but1 || but2 || but3) == 0) Win_alarm_buts[1]->label = "";

   for(i = 0; i < 3; ++i) {
      Tlk_set(Win_alarm_buts[i], Win_alarm_buts[i]->label ? TLK_SHOW : TLK_UNSHOW, 0, 0);
   }

   if(Dsp_color) {
      Win_3d(Win_alarm_win->pixmap, Dspgc_25, Dspgc_75, Dspgc_50, 2, 2, 
      WIN_ALARM_WIDTH  * Dsp_rw[Win_alarm_win->scale] - 5, 
      WIN_ALARM_HEIGTH * Dsp_rh[Win_alarm_win->scale] - 5);
   } else {
      Win_fb(Win_alarm_win, Dspgc_clr, 0, 0, 500, 500, 0, 0);
   }
      
   if(hl) {
      while(*hl == ' ') ++hl;
      sprintf(Win_buf,"%s: %s", Dsp_prog, hl);
   } else {
      strcpy(Win_buf, Dsp_prog);
   }
      
   x = WIN_ALARM_WIDTH/2 - strlen(Win_buf)/2;
   if(x < 4) x = 4;

   Win_bs(Win_alarm_win, Dspgc, 1, x, 0, 0, Win_buf);
          
   x = (Dsp_rh[Win_alarm_win->scale] * 1.5) - 13;
   
   XCopyArea(Dsp, Win_stop_pix, Win_alarm_win->pixmap, Dspgc, 0, 0, 25, 25, x, x);
   
   Win_dl(Win_alarm_win, Dspgc,    3, 1, 3, WIN_ALARM_WIDTH - 2, 0,  0);
   Win_dl(Win_alarm_win, Dspgc_25, 3, 1, 3, WIN_ALARM_WIDTH - 2, 0, -1);

   for(i = 0; i < 5; ++i) {
      if(lines[i] == 0) continue;
      if(strlen(lines[i]) > WIN_ALARM_WIDTH - 2)
         lines[i+ WIN_ALARM_WIDTH - 2] = '\0';
      Win_ds(Win_alarm_win, Dspgc, WIN_ALARM_FIRST_ROW + i,
             WIN_ALARM_WIDTH/2 - strlen(lines[i])/2, 0, 0, lines[i]);
   }

   Win_lastx -= Win_alarm_win->w / WIN_ALARM_WIDTH * (but2 ? 28 : 42);
   Win_lasty -= Win_alarm_win->h / WIN_ALARM_HEIGTH * 11;

   if(Win_lastx < 0) Win_lastx = 2;
   if(Win_lasty < 0) Win_lasty = 2;

   if(Win_lastx > DisplayWidth(Dsp, Dsp_screen)  - Win_alarm_win->w)
      Win_lastx = DisplayWidth(Dsp, Dsp_screen)  - Win_alarm_win->w - 6;
   if(Win_lasty > DisplayHeight(Dsp, Dsp_screen) - Win_alarm_win->h)
      Win_lasty = DisplayHeight(Dsp, Dsp_screen) - Win_alarm_win->h - 6;

   XMoveWindow(Dsp, Win_alarm_win->window, Win_lastx, Win_lasty);
   XMapRaised(Dsp, Win_alarm_win->window);

   XGrabPointer(Dsp, Win_alarm_win->window, TRUE,
           ButtonPressMask | ButtonReleaseMask |
           EnterWindowMask | LeaveWindowMask,
           GrabModeAsync, GrabModeAsync, None,
           Dspcur[(int)DSPCUR_ALR], CurrentTime);

   XGrabKeyboard(Dsp, Win_alarm_win->window, TRUE, GrabModeAsync, GrabModeAsync, CurrentTime);

   Win_alarm_active = TRUE;
   Win_alarm_but = WIN_ALARM_NO_BUT;

   But_xbutton_pressed = FALSE;
   Win_loop(0);
   Win_alarm_active = FALSE;

   XUngrabPointer(Dsp, CurrentTime);
   XUngrabKeyboard(Dsp, CurrentTime);
   XUnmapWindow(Dsp, Win_alarm_win->window);

   XFlush(Dsp);
   return Win_alarm_but;
}

void Win_alarm_proc(But but)   {
   Win_alarm_but = (Win_alarm_but_t) WIN_ALARM_BUT(but->x / Dsp_rw[but->scale]);
}

void Win_draw(Win win, GC gc, int r1, int c1, int r2, int c2, int xoff, int yoff, int what) {
   int             x, y;
   int             w, h;
   x = c1 * Dsp_rw[win->scale] + xoff - 1;
   y = r1 * Dsp_rh[win->scale] + yoff - 1;
   if(what == 0) {
      ++x;
      w = c2 * Dsp_rw[win->scale] + xoff;
      h = r2 * Dsp_rh[win->scale] + yoff - 1;
   } else {
      w = c2 * Dsp_rw[win->scale] + 1;
      h = r2 * Dsp_rh[win->scale];
   }
   if(Win_alarm_win) {
      switch(what) {
         case 0: /* dl */
            XDrawLine(Dsp, win->window, gc, x, y, w, h);
            break;
         case 1: /* db */
            XDrawRectangle(Dsp, win->window, gc, x, y, w, h);
            break;
         case 2: /* fb */
            XFillRectangle(Dsp, win->window, gc, x, y, w, h);
            break;
      }
   }
   if(win->pixmap) {
      switch(what) {
         case 0: /* dl */
            XDrawLine(Dsp, win->pixmap, gc, x, y, w, h);
            break;
         case 1: /* db */
            XDrawRectangle(Dsp, win->pixmap, gc, x, y, w, h);
            break;
         case 2: /* fb */
            XFillRectangle(Dsp, win->pixmap, gc, x, y, w, h);
            break;
      }
   }
}         

void Win_dl(Win win, GC gc, int r1, int c1, int r2, int c2, int xoff, int yoff) {
   Win_draw(win, gc, r1, c1, r2, c2, xoff, yoff, 0);
}

void Win_db(Win win, GC gc, int row, int col, int rows, int cols, int xoff, int yoff) {
   Win_draw(win, gc, row, col, rows, cols, xoff, yoff, 1);
}

void Win_fb(Win win, GC gc, int row, int col, int rows, int cols, int xoff, int yoff) {
   Win_draw(win, gc, row, col, rows, cols, xoff, yoff, 2);
}

void Win_text(Win win, GC gc, int row, int col, int xoff, int yoff, unsigned char *str, unsigned int len, int what) {
   int             x, y;

   if(len == 0) return;

   x =   col * Dsp_rw[win->scale] + xoff;
   y = ++row * Dsp_rh[win->scale] - Dsp_fd[win->scale] - 2 + yoff;

   if(Dsp_ul[win->scale] == FALSE) {
      register int cnt = len;
      while(cnt--) {
         switch(str[cnt]) {
            case TLK_Adiaeresis: Win_uml[cnt] = 'A';      break;
            case TLK_Odiaeresis: Win_uml[cnt] = 'O';      break;
            case TLK_Udiaeresis: Win_uml[cnt] = 'U';      break;
            case TLK_adiaeresis: Win_uml[cnt] = 'a';      break;
            case TLK_odiaeresis: Win_uml[cnt] = 'o';      break;
            case TLK_udiaeresis: Win_uml[cnt] = 'u';      break;
            case TLK_ssharp    : Win_uml[cnt] = 'B';      break;
            default:             Win_uml[cnt] = str[cnt]; break;
         }
      }
      str = (unsigned char*) Win_uml;
   }
   if(Win_alarm_win) {
      switch(what) {
         case 0: /* ft */
            XDrawImageString(Dsp, win->window, gc, x, y, (char*)str, len);
            break;
         case 2: /* bt */
            XDrawString(Dsp, win->window, gc, x + 1, y, (char*)str, len);
         case 1: /* dt */
            XDrawString(Dsp, win->window, gc, x, y, (char*)str, len);
         };
   }
   if(win->pixmap) {
      switch(what) {
         case 0: /* ft */
            XDrawImageString(Dsp, win->pixmap, gc, x, y, (char*)str, len);
            break;
         case 2: /* bt */
            XDrawString(Dsp, win->pixmap, gc, x + 1, y, (char*)str, len);
         case 1: /* dt */
            XDrawString(Dsp, win->pixmap, gc, x, y, (char*)str, len);
      };
   }
}

void Win_bs(Win win, GC gc, int row, int col, int xoff, int yoff, char *str) {
   Win_text(win, gc, row, col, xoff, yoff, (unsigned char*) str, strlen(str), 2);
}

void Win_ds(Win win, GC gc, int row, int col, int xoff, int yoff, char *str) {
   Win_text(win, gc, row, col, xoff, yoff, (unsigned char*) str, strlen(str), 1);
}

void Win_fs(Win win, GC gc, int row, int col, int xoff, int yoff, char *str) {
   Win_text(win, gc, row, col, xoff, yoff, (unsigned char*) str, strlen(str), 0);
}

void Win_bt(Win win, GC gc, int row, int col, int xoff, int yoff, char *str, int len) {
   Win_text(win, gc, row, col, xoff, yoff, (unsigned char*) str, len, 2);
}

void Win_dt(Win win, GC gc, int row, int col, int xoff, int yoff, char *str, int len) {
   Win_text(win, gc, row, col, xoff, yoff, (unsigned char*) str, len, 1);
}

void Win_ft(Win win, GC gc, int row, int col, int xoff, int yoff, char *str, int len) {
   Win_text(win, gc, row, col, xoff, yoff, (unsigned char*) str, len, 0);
}

void Win_set_middle_array(Win win, int row, int col, int len) {
   if(win->type == WIN_BAS_WIN) while (len) *(win->middle + (win->cols * row + col + --len)) = TRUE;
}

void Win_3d(Window win, GC l, GC r, GC m, int x, int y, int w, int h) {
   XDrawLine(Dsp, win, l, x-2, y-2, x-2, y+h+2);  /*Links*/
   XDrawLine(Dsp, win, l, x-1, y-1, x-1, y+h);  
   XDrawLine(Dsp, win, l, x-2, y-2, x+w+2, y-2);  /*Oben*/
   XDrawLine(Dsp, win, l, x-1, y-1, x+w, y-1);
   XDrawLine(Dsp, win, r, x-1, y+h+1, x+w+1, y+h+1); /*Unten*/
   XDrawLine(Dsp, win, r, x-2, y+h+2, x+w+2, y+h+2);
   XDrawLine(Dsp, win, r, x+w+1, y-1, x+w+1, y+h+1);  /*Rechts*/
   XDrawLine(Dsp, win, r, x+w+2, y-2, x+w+2, y+h+2);
   XFillRectangle(Dsp, win, m, x, y, w+1, h+1);
}

void *Win_malloc(int size) {
   return Win_realloc(0, size);
}

void *Win_realloc(void *optr, int size) {
   void *nptr = (optr) ? (void *) realloc(optr, ++size) : (void *) malloc(++size);
   if(nptr == 0) {
      fprintf(stderr, "%s: re/malloc(%d) failed: %s.\n", Dsp_prog, size, strerror(errno));
      Win_exit();
   }
   return nptr;
}

void Win_alarm_clock() {
   if(Win_lock == FALSE && Mnu_act == -1) {
      if(Win_clock_proc) {    
         (void) (*Win_clock_proc) ();
      }
   }
   signal(SIGALRM, Win_alarm_clock);
   alarm(Win_alarm_reload / 1000);
}

void Win_disable_timer() {
   signal(SIGALRM, SIG_IGN);
}

void Win_enable_timer() {
   signal(SIGALRM, Win_alarm_clock);
   alarm(Win_alarm_init / 1000);
}

void Win_set_itimer(int init, int reload, void (*proc)()) {
   Win_alarm_init   = init;
   Win_alarm_reload = reload;
   Win_clock_proc   = proc;
   Win_enable_timer();
}

int Win_error(Display *d, XErrorEvent *xe) {
   fprintf(stderr, "%s: Internal X11-error, display = %s\n", Dsp_prog, DisplayString(d));
   fprintf(stderr, "%s: id = %ld error = %d req = %d\n", Dsp_prog, xe->resourceid, xe->error_code, xe->request_code);
   switch (xe->error_code) {
      case 1:
      case 2:
      case 3:
      case 16:
         Win_exit();
   }
   return TRUE;
}

void Win_exit() {
   fprintf(stderr, "%s: Program terminated.\n", Dsp_prog);
   Win_quit();
}

void Win_quit() {
   Txt   ntxt, txt = Txts;
   Win   nwin, win;
   int   factor;

   while(txt) {
      ntxt = txt->next;
      free(txt);
      txt = ntxt;
   }
   
   for(factor = 0; factor < WIN_FACTOR; ++factor) {
      win = Wins[factor];
      while(win) {
         switch(win->type) {
            case WIN_LST_WIN:
               Dbk_destroy(win->data);
               break;
            case WIN_BAS_WIN:
            case WIN_POP_WIN:
            case WIN_SUB_WIN:
               free(win->data);
               free(win->text);
               free(win->middle);
            default:
               break;
         }
         nwin = win->next;
         free(win);
         win = nwin;
      }
   }
   if(Win_io) free(Win_io);
   if(Dsp) {
      XFlush(Dsp);
      XCloseDisplay(Dsp);
      exit(0);
   } else {
      exit(1);
   }
}

#endif            /* _Win_c */
